"""
mod_utils.py

Module utilities functions.
"""

from functools import wraps


#################################################
#################################################


def check_dependency(dep, name):
    """Decorator that checks if an optional dependency is available.

    Parameters
    ----------
    dep : module or False
        Module, if successfully imported, or boolean (False) if not.
    name : str
        Full name of the module, to be printed in message.

    Returns
    -------
    wrap : callable
        The decorated function.

    Raises
    ------
    ImportError
        If the requested dependency is not available.
    """

    def wrap(func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if not dep:
                raise ImportError("Optional dependency " + name + \
                                  " is required for this functionality.")
            return func(*args, **kwargs)
        return wrapped_func
    return wrap


DOCSTRING_SECTIONS = ['Parameters', 'Returns', 'Yields', 'Raises',
                      'Warns', 'Examples', 'References', 'Notes',
                      'Attributes', 'Methods']



"""
crosscorrelograms.py

Modules for generating crosscorrelograms.
"""

import numpy as np
from joblib import Parallel, delayed

from synapticonn.postprocessing.correlogram_utils import make_bins


##########################################################
##########################################################


def compute_crosscorrelogram(spike_times, spike_pairs, bin_size_t, max_lag_t):
    """ Compute the cross-correlogram between all pairs of spike trains.

    Parameters
    ----------
    spike_times : dict
        Dictionary containing spike times for each unit. Indexed by unit ID.
    spike_pairs : list
        List of tuples containing the unit IDs of the spike trains to compare.
    bin_size_t : float
        Bin size of the cross-correlogram.
    max_lag_t : float
        Maximum lag to compute the cross-correlogram.

    Returns
    -------
    cross_correlograms_data : dict
        Dictionary containing cross-correlograms and bins for all pairs
        of spike trains. Indexed by unit ID pairs

    Notes
    -----
    Parallel processing is performed across all available cores.
    """

    def _process_pair(pair):
        pair_1, pair_2 = pair
        cross_corr, bins = compute_crosscorrelogram_dual_spiketrains(
            spike_times[pair_1], spike_times[pair_2], bin_size_t, max_lag_t
        )
        return (pair_1, pair_2), cross_corr, bins

    results = Parallel(n_jobs=-1)(delayed(_process_pair)(pair) for pair in spike_pairs)

    cross_corr_dict = {key: cross_corr for key, cross_corr, _ in results}
    bins_dict = {key: bins for key, _, bins in results}

    crosscorrelogram_data = {'cross_correllations': cross_corr_dict, 'bins': bins_dict}

    return crosscorrelogram_data


def compute_crosscorrelogram_dual_spiketrains(spike_train_1, spike_train_2, bin_size_t, max_lag_t):
    """ Compute the cross-correlogram between two spike trains.

    Parameters
    ----------
    spike_train_1 : array_like
        The spike times of the first spike train.
    spike_train_2 : array_like
        The spike times of the second spike train.
    bin_size_t : float
        The size of the bins in which to bin the time differences (in milliseconds).
    max_lag_t : float
        The maximum lag to consider (in milliseconds).

    Returns
    -------
    cross_corr : array_like
        The cross-correlogram.
    """

    # convert to numpy arrays
    spike_train_1 = np.array(spike_train_1)
    spike_train_2 = np.array(spike_train_2)

    time_diffs = spike_train_2[:, np.newaxis] - spike_train_1

    mask = (time_diffs >= -max_lag_t) & (time_diffs <= max_lag_t)
    valid_diffs = time_diffs[mask]

    bins = make_bins(max_lag_t, bin_size_t)
    cross_corr, _ = np.histogram(valid_diffs, bins)

    return cross_corr, bins




""" spike_times.py

Object for handling spike time data.
"""

import warnings
import logging

import numpy as np
import pandas as pd

from synapticonn.utils.errors import SpikeTimesError, DataError, RecordingLengthError, SamplingRateError
from synapticonn.utils.attribute_checks import requires_arguments
from synapticonn.quality_metrics import compute_isi_violations, compute_presence_ratio, compute_firing_rates
from synapticonn.core.core_utils import setup_log
from synapticonn.core.info import get_unit_time_types, get_quality_metric_keys
from synapticonn.utils.warnings import custom_formatwarning


###############################################################################
###############################################################################

warnings.formatwarning = custom_formatwarning

###############################################################################
###############################################################################


class SpikeManager():
    """
    Base class for managing and processing spike time data.

    This class provides methods for validating, organizing, and analyzing
    spike time data for neuroscience studies. It facilitates quality control
    metrics computation, spike unit filtering, and retrieval of relevant spike
    unit information.

    Parameters
    ----------
    spike_times : dict
        Dictionary containing spike times for each unit, indexed by unit ID.
    time_unit : str
        Unit of time for the spike times. Default is 'ms'. Options include 's', 'ms'.
    srate : float
        Sampling rate of the recording in Hz.
    recording_length_t : float
        Length of the recording in the specified time unit.
    spike_id_type : type
        Data type of the spike unit ID. Default is `int` or `str`.

    Attributes
    ----------
    spike_times : dict
        Validated dictionary of spike times, indexed by unit ID.
    time_unit : str
        The time unit used for spike times ('ms' or 's').
    recording_length_t : float
        The duration of the recording in the specified time unit.
    spike_id_type : type
        Data type of spike unit identifiers.
    srate : float
        Sampling rate of the recording in Hz.
    spike_unit_filtering : bool
        Tracks whether spike units have been filtered.

    Notes
    -----
    The SpikeManager object is used to manage spike time data.
    This object is used to store and process spike times for
    each unit in the recording. The object provides methods
    for computing quality metrics, filtering spike units,
    and reporting spike unit information.

    The SpikeManager object is initialized with the spike times,
    time unit, sampling rate, recording length, and spike ID type.
    The spike times are stored in a dictionary, indexed by the unit ID.

    The SpikeManager object provides methods for computing quality metrics
    for each unit, filtering spike units based on quality metrics,
    and reporting spike unit information. These methods are used to
    analyze and process the spike time data.
    """


    # spike unit filtering flag to track if units have been filtered
    spike_unit_filtering = False


    def __init__(self, spike_times: dict = None,
                 time_unit: str = 'ms',
                 srate: float = None,
                 recording_length_t: float = None,
                 spike_id_type: type = int or str):
        """ Initialize the spike manager. """

        # prepare the spike time data
            # set the spike times, time unit, recording length, and spike ID type
        self.spike_times, self.time_unit, self.recording_length_t, self.spike_id_type, self.srate = \
            self._prepare_spiketime_data(spike_times, time_unit, recording_length_t, spike_id_type, srate)


    def _reset_spike_time_data(self):
        """ Reset the spike times data. """

        self.spike_times = None
        self.recording_length_t = None
        self.srate = None
        self.spike_id_type = None
        self.time_unit = None
        self.spike_unit_filtering = False  # reset the spike unit filtering flag


    def add_spike_time_data(self, spike_times: dict = None,
                            recording_length_t: float = None,
                            time_unit: str = 'ms',
                            srate: float = None,
                            spike_id_type: type = int or str):
        """ Add spike time data to the SpikeManager object.

        Parameters
        ----------
        spike_times : dict
            Dictionary containing spike times for each unit.
            Indexed by unit ID.
        recording_length_t : float
            Length of the recording in time units.
        time_unit : str
            Unit of time for the spike times. Default is 'ms'.
            Options include 's', 'ms'.
        srate : float
            Sampling rate of the recording in Hz.
        spike_id_type : type
            Type of the spike unit ID. Default is int or str.
        """

        # if any data is already present, then clear
        # and the results to ensure object consistency
        self._reset_spike_time_data()

        # prepare the spike time data
        self.spike_times, self.time_unit, self.recording_length_t, self.spike_id_type, self.srate = \
            self._prepare_spiketime_data(spike_times, time_unit, recording_length_t, spike_id_type, srate)


    def report_spike_units(self):
        """ Report the spike units.

        Returns
        -------
        spk_unit_summary : dict
            Dictionary containing the spike unit summary.
            Includes the unit ID, number of spikes, and
            firing rate in Hz.

        Notes
        -----
        The spike unit summary is computed for each unit
        in the spike_times dictionary. Firing rates
        are calculated based on the total number of spikes
        and the recording length. If the time unit is in seconds,
        the firing rate is converted to Hz.
        """

        labels = self.spike_unit_labels()

        # number of spikes for each unit
        n_spks = [len(self.spike_times[label]) for label in labels]

        # calculate firing rates for each unit
            # convert to Hz if time unit is in seconds
        if self.time_unit == 's':
            firing_rates = [len(self.spike_times[label]) / self.recording_length_t for label in labels]
        elif self.time_unit == 'ms':
            firing_rates = [len(self.spike_times[label]) / self.recording_length_t * 1000 for label in labels]
        else:
            raise ValueError(f"Invalid time_unit: {self.time_unit}. Must be 's' or 'ms'.")

        spk_unit_summary = {'unit_id': labels,
                            'n_spikes': n_spks,
                            'firing_rate_hz': firing_rates}

        return spk_unit_summary


    def spike_unit_labels(self):
        """ Retrieve the spike unit labels. """

        return list(self.spike_times.keys())


    def get_spike_times_for_units(self, spike_units_to_collect: list = None) -> dict:
        """ Retrieve spike times for the selected units.

        Parameters
        ----------
        spike_units_to_collect : list
            List of spike units to collect.

        Returns
        -------
        filtered_spike_times : dict
            Filtered dictionary containing spike times for the selected units.

        Notes
        -----
        By default, this method returns all spike times if no units are specified.
        If specific units are specified, then the method will return the spike times
        for those units only.
        """

        if not isinstance(spike_units_to_collect, list):
            raise TypeError("Expected list of spike units. "
                            f"Got {type(spike_units_to_collect)} instead.")

        # return all spike times if no units are specified
        if spike_units_to_collect is None:
            print("No units specified. Returning all spike times.")
            filtered_spike_times = self.spike_times

        # return the selected units
        else:
            try:
                filtered_spike_times = {key: self.spike_times[key] for key in spike_units_to_collect}
            except KeyError as e:
                raise SpikeTimesError(f"Unit {e} not found in the spike times dictionary.")

        return filtered_spike_times


    def spike_unit_quality(self,
                           isi_threshold_ms=1.5,
                           min_isi_ms=0,
                           presence_ratio_bin_duration_sec=60,
                           presence_ratio_mean_fr_ratio_thresh=0.0) -> pd.DataFrame:
        """ Compute spike isolation quality metrics.

        Parameters
        ----------
        isi_threshold_ms : float
            Threshold for the interspike interval (ISI) violations, in milliseconds.
        min_isi_ms : float
            Minimum ISI value, in milliseconds.
        presence_ratio_bin_duration_sec : float
            Duration of each bin for the presence ratio, in seconds.
        presence_ratio_mean_fr_ratio_thresh : float
            Minimum mean firing rate ratio threshold for the presence ratio.
            This is the minimum mean firing rate that must be present in a bin
            for the unit to be considered "present" in that bin.
            By default, this is set to 0.0. This means that the unit must have
            at least one spike in each bin to be considered "present."
        time_unit : str
            Unit of time for the spike times. Default is 'ms'.
            This is used to compute the features in the quality metrics.

        Returns
        -------
        quality_metrics : pd.DataFrame
            DataFrame containing the quality metrics for each spike unit.

        Notes
        -----
        Quality metrics include:
        - isi_violations_ratio: Fraction of ISIs that violate the threshold.
        - isi_violations_count: Number of ISIs that violate the threshold.
        - isi_violations_of_total_spikes: Fraction of ISIs that violate the threshold out of total spikes.
        - presence_ratio: Fraction of time during a session in which a unit is spiking.
        - mean_firing_rate: Mean firing rate of the unit.
        - recording_length_sec: Length of the recording in seconds.
        - n_spikes: Number of spikes for the unit.

        These are computed for each spike unit in the spike_times dictionary.

        For further information on the quality metric calculations,
        see the respective functions in the quality_metrics module.
        """

        quality_metrics = {}
        for key, spks in self.spike_times.items():

            # isi violations
            isi_violations = compute_isi_violations(spks,
                                                    self.recording_length_t,
                                                    isi_threshold_ms,
                                                    min_isi_ms,
                                                    self.time_unit)

            # presence ratio
            presence_ratio = compute_presence_ratio(spks,
                                                    self.recording_length_t,
                                                    self.time_unit,
                                                    presence_ratio_bin_duration_sec,
                                                    presence_ratio_mean_fr_ratio_thresh,
                                                    self.srate)

            # unit firing rates
            firing_rates = compute_firing_rates(spks,
                                                self.recording_length_t,
                                                self.time_unit)

            quality_metrics[key] = isi_violations
            quality_metrics[key].update(presence_ratio)
            quality_metrics[key].update(firing_rates)

        return pd.DataFrame(quality_metrics).T


    def filter_spike_units(self, quality_metrics: pd.DataFrame,
                           query: str = None,
                           log: bool = False,
                           overwrite: bool = False) -> pd.DataFrame:
        """ Filter spike units based on quality metrics.

        Parameters
        ----------
        quality_metrics : pd.DataFrame
            DataFrame containing the quality metrics for each spike unit.
            This is the dataframe outputted from the spike_unit_quality method
            and will be used to filter spike units.
        query : str
            Query to filter spike units based on the quality metrics.
            This query should be a valid pandas query
        log : bool
            Whether to log the filtered spike units. Default is False.
        overwrite : bool
            Whether to overwrite the existing spike_times dictionary with the filtered units.
            Default is False.

        Returns
        -------
        filtered_units_df : pd.DataFrame
            DataFrame containing the filtered spike units based on the query.

        Log
        ---
        If log is True, the method will log the removed spike units
        based on the query. The log will contain the unit ID and the
        query used to filter the units.

        The log file will be saved in the 'removed_spike_units' folder
        in the current working directory. The log file will be named
        'low_quality_units_removed.log'.
        """

        if quality_metrics is None:
            raise DataError("Quality metrics DataFrame is missing. "
                            "Please run the spike_unit_quality method.")

        # check the query
        assert isinstance(query, str), f"Query must be a string. Got {type(query)} instead."
        # check the quality metrics type
        assert isinstance(quality_metrics, pd.DataFrame), "Quality metrics must be a DataFrame. \
                                                            Got {type(quality_metrics)} instead."

        # check if spike units have already been filtered
        if SpikeManager.spike_unit_filtering:
            if not overwrite:
                msg = ("Spike units have already been filtered. Please re-initialize the object "
                       "or 'set_spike_times' to set the spike_times dict for re-filtering. "
                       "If this was intentional, please set the 'overwrite' parameter to True.")
                warnings.warn(msg)
            if overwrite:
                SpikeManager.spike_unit_filtering = False

        if not set(get_quality_metric_keys()).issubset(quality_metrics.columns):
            msg = ("Quality metrics DataFrame is missing required columns. "
                   f"Required columns: {get_quality_metric_keys()}. "
                   "Please run the spike_unit_quality method.")
            raise DataError(msg)

        # filter units based on query
        if query is not None:
            filtered_units_df = quality_metrics.query(query)
        else:
            # if no query, return the original dataframe
            filtered_units_df = quality_metrics

        # remove filtered units from the spike times dictionary
        self.spike_times = {key: self.spike_times[key] for key in filtered_units_df.index}

        # if log, track removed units
        if log:

            setup_log(log_folder_name='removed_spike_units',
                      log_fname='low_quality_units_removed.log')

            removed_units = quality_metrics[~quality_metrics.index.isin(filtered_units_df.index)]

            for key, row in removed_units.iterrows():
                log_msg = f'unit_id: {key} - unit removed from original dataframe with query {query}'
                logging.info(log_msg)

        SpikeManager.spike_unit_filtering = True

        return filtered_units_df


    @requires_arguments('spike_id_type','srate', 'spike_times',
                        'recording_length_t', 'time_unit')
    def _prepare_spiketime_data(self, spike_times: dict = None,
                                time_unit: str = None,
                                recording_length_t: float = None,
                                spike_id_type: type = None,
                                srate: type = float or int):
        """ Prepare the spike time data.

        Parameters
        ----------
        spike_times : dict
            Dictionary containing spike times for each unit.
            Indexed by unit ID.
        srate : float or int
            Sampling rate of the recording in Hz.
        time_unit : str
            Unit of time for the spike times. Default is 'ms'.
            Options include 's', 'ms'.
        recording_length_t : float
            Length of the recording in time units.
        spike_id_type : type
            Type of the spike unit ID. Default is int or str.

        Returns
        -------
        Returns the validated spike times, time unit, recording length, and spike ID type.
        """

        ## sampling rate checks - run these checks on the sampling rate input

        if not isinstance(srate, (float, int)):
            raise SamplingRateError("Sampling rate must be a float or int.")
        if srate <= 0:
            raise SamplingRateError("Sampling rate must be greater than 0.")
        if srate < 5000 or srate > 30_000:
            warnings.warn("Sampling rate is outside the typical range of 5000-30,000 Hz. "
                          "Please verify the sampling rate used to perform spike unit extractions. "
                          "Please taken into account the Nyquist frequency when analyzing the data. "
                          "If this is intentional, please ignore this warning.")

        ## spike time checks - run these checks on the spike times dict input

        # check that spike_times is a dictionary
            # see documentation for more information on the expected format
        if not isinstance(spike_times, dict):
            raise SpikeTimesError("Spike times must be a dictionary. Each key should be the unit ID.")

        # check that all keys in spike_times match the expected type
        if not all(isinstance(key, spike_id_type) for key in spike_times.keys()):
            raise SpikeTimesError(
                f"All keys in spike_times must be of type '{spike_id_type.__name__}'. "
                "Please verify the unit IDs or update the spike_id_type."
            )

        # validate the spike times for each unit
        for unit_id, spks in spike_times.items():
            # check if spike times array is empty
            if len(spks) == 0:
                raise SpikeTimesError(f"Spike times for unit {unit_id} are empty. Please check the data.")

            # check for non-negative spike times
            if not np.all(spks >= 0):
                raise SpikeTimesError(f"Spike times for unit {unit_id} must be non-negative.")

            # check that each spike time is a float
            if not np.issubdtype(spks.dtype, np.floating):
                raise SpikeTimesError(
                    f"Spike times for unit {unit_id} must be of type 'float'. "
                    f"Found type '{spks.dtype}' instead."
                )

            # check if any nan or inf values in spike times
            if np.any(np.isnan(spks)) or np.any(np.isinf(spks)):
                raise SpikeTimesError(f"Spike times for unit {unit_id} contain NaN or Inf values.")

        # check the time unit type
        time_unit = self._time_unit_check(time_unit)

        # ensure spike times do not exceed recording length
            # if so, probably an error in the recording length, spike times or time unit
        for unit_id, spks in spike_times.items():
            if np.max(spks) > recording_length_t:
                msg = (f"Spike times for unit {unit_id} exceed the recording length. "
                       f"Max spike time: {np.max(spks)}, Recording length: {recording_length_t}. "
                       f"Check that the recording length is correct and in {time_unit}.")
                raise RecordingLengthError(msg)

        # check the time unit matches the recording length
            # raise a warning if the spike times exceed the recording length
        max_spk_time = np.max([np.max(spks) for spks in spike_times.values()])
        if max_spk_time < 0.5*recording_length_t:
            msg = ("Unit is firing across less than 50% of the recording length. "
                   "This may lead to unexpected results. Please check the spike times and recording length "
                   "passed to the SpikeManager object. If this is intentional, please ignore this warning.")
            warnings.warn(msg)

        return spike_times, time_unit, recording_length_t, spike_id_type, srate


    def _time_unit_check(self, time_unit: str = None):
        """ Check the time unit. """

        if time_unit not in get_unit_time_types():
            raise TypeError(
                f"Time unit must be in {get_unit_time_types()}. "
                f"Got {time_unit} instead."
                )

        return time_unit


""" ccg_connection_type.py

Modules to categorize connection types
based on synaptic strengths computed
using the cross-correlogram (CCG) method.
"""

import warnings
import numpy as np

from synapticonn.utils.errors import ConnectionTypeError


##########################################################
##########################################################


def get_putative_connection_type(synaptic_strength, threshold=None):
    """ Calculate putative connection type based on synaptic strengths.

    Parameters
    ----------
    synaptic_strength : float
        Synaptic strength value.
    threshold : float, optional
        Threshold value for categorizing connection types. Default is None.

    References
    ----------
    [1] STAR Protoc. 2024 Jun 21;5(2):103035. doi: 10.1016/j.xpro.2024.103035. Epub 2024 Apr 27
    """

    assert isinstance(synaptic_strength, (int, float, np.number)), "Synaptic strength must be a number."
    assert threshold is None or isinstance(threshold, (int, float, np.number)), "Threshold must be a number."

    if threshold > 0:
        return _get_exc_putative_connection_type(synaptic_strength, threshold)
    elif threshold < 0:
        return _get_inh_putative_connection_type(synaptic_strength, threshold)
    else:
        raise ConnectionTypeError("Threshold must be non-zero.")


def _get_exc_putative_connection_type(synaptic_strength, threshold=5):
    """ Categorize excitatory connection types based on synaptic strengths. """

    if threshold < 5:
        warnings.warn("Threshold is < 5. Recommended to use a threshold of >= 5 for excitatory connections.")

    if synaptic_strength >= threshold:
        label = "excitatory monosynaptic"
    elif synaptic_strength < threshold:
        label = 'undefined'

    return {'putative_exc_connection_type': label}


def _get_inh_putative_connection_type(synaptic_strength, threshold=5):
    """ Categorize inhibitory connection types based on synaptic strengths. """

    if threshold > -5:
        warnings.warn("Threshold is > -5. Recommended to use a threshold of <= -5 for inhibitory connections.")

    if synaptic_strength <= threshold:
        label = "inhibitory monosynaptic"
    elif synaptic_strength > threshold:
        label = 'undefined'

    return {'putative_inh_connection_type': label}


""" mono_connection_peak.py 

Modules for computing the features of cross-correlogram (CCG) peaks.
"""

import numpy as np 


##################################################################
##################################################################


def compute_peak_amp(ccg):
    """ Compute the peak amplitude (count per bin) of the CCG.

    Parameters
    ----------
    ccg : array
        Cross-correlogram.

    Returns
    -------
    output : float
        Amplitude of the peak of the CCG
    """

    # find the peak of the cross-correlogram
    peak = np.max(ccg)

    return {'ccg_peak_count_per_bin': peak}


"""Sub-module for plot functions."""

from .acg import plot_acg
from .ccg import plot_ccg
from .synaptic_strength_calc import plot_ccg_synaptic_strength
from .spiketrain import plot_spiketrain
from .plot_utils import acg_ax, check_spktime_ax_length, ccg_ax, spktrain_ax
from .style import apply_plot_style


""" mono_connection_lateny.py 

Modules for computing the timing of cross-correlogram (CCG) peaks.
"""

import numpy as np 


##################################################################
##################################################################


def compute_peak_latency(ccg, bin_size_t):
    """ Compute the peak latency of the CCG.

    Parameters
    ----------
    ccg : array
        Cross-correlogram.
    bin_size_t : float
        Bin size of the cross-correlogram.

    Returns
    -------
    output : float
        Timing of the peak of the CCG
    """

    # find the peak of the cross-correlogram
    peak = np.argmax(ccg)

    # convert the peak to time
    peak_time = (peak - len(ccg) // 2) * bin_size_t

    return {'ccg_peak_time_ms': peak_time}


""" isi_violations.py.

Modules for computing interspike interval (ISI) violations in a spike train.
"""


import warnings
import numpy as np


####################################################
####################################################


def compute_isi_violations(spike_train,
                           recording_length_t,
                           isi_threshold_ms=1.5,
                           min_isi_ms=0,
                           time_unit='ms'):
    """Compute the number of interspike interval (ISI) violations in a spike train.

    Parameters
    ----------
    spike_train : numpy.ndarray (1D)
        Spike train.
    recording_length_t : float
        Length of the recording.
    time_unit : str
        Time unit of the recording. Options are 'ms' or 's'.
    isi_threshold_ms : float
        Minimum interspike interval, in ms.
        This is the minimum time that must elapse between two spikes.
    min_isi_ms : float
        Minimum possible interspike interval in ms.
        This is the artifical refractory period enforced by the
        recording system or the spike sorting algorithm.

    Notes
    -----
    All times are in milliseconds (ms) by default.
    If not, the recording length and spike times will be converted to milliseconds.

    References
    ----------
    Based on hte metrics orginally implemented in the SpikeInterface package
    (https://github.com/SpikeInterface/spikeinterface/blob/main/src/spikeinterface/qualitymetrics/misc_metrics.py).
    This was based on metrics originally implemented in Ultra Mega Sort [UMS]_.

    This implementation is based on one of the original implementations written in Matlab by Nick Steinmetz
    (https://github.com/cortex-lab/sortingQuality) and converted to Python by Daniel Denman.

    Documentation / resources
    --------------------------
    For future documentation on isi violations, see:
    https://allensdk.readthedocs.io/en/latest/_static/examples/nb/ecephys_quality_metrics.html#ISI-violations

    This documentation by Allen Brain provided information on what thresholds may be considered 
    acceptable for ISI violations.
    """

    isi_violations = {'isi_violations_ratio': np.nan,
                      'isi_violations_count': np.nan,
                      'isi_violations_of_total_spikes': np.nan}

    # if the time unit is in seconds, warn the user
    # that the ISI violations will be skipped and
    # that the time unit should be converted to milliseconds
    if time_unit == 's':
        warnings.warn("The time unit is set to seconds. Spike times "
                      "will be converted to milliseconds to calculate ISI violations.")
        spike_times_ms = spike_train * 1000
        recording_length_ms = recording_length_t * 1000
    elif time_unit == 'ms':
        spike_times_ms = spike_train
        recording_length_ms = recording_length_t
    else:
        raise ValueError("Time unit must be 'ms' or 's'.")

    # warn the user if the ISI threshold is
    # greater than the refractory period
    if (isi_threshold_ms > 1.5):
        warnings.warn("The ISI threshold is set to a value greater than the "
                      "refractory period of most neurons.")

    isi_violations_count = {}
    isi_violations_ratio = {}

    isis = np.diff(spike_times_ms)
    num_spikes = len(spike_times_ms)
    num_violations = np.sum(isis < isi_threshold_ms)

    violation_time = 2 * num_spikes * (isi_threshold_ms - min_isi_ms)

    if num_spikes > 0:
        total_rate = num_spikes / (recording_length_ms / 1000)
        violation_rate = num_violations / violation_time
        isi_violations_ratio = violation_rate / total_rate
        isi_violations_count = num_violations
        isi_violations_of_total = num_violations / num_spikes

    isi_violations['isi_violations_ratio'] = isi_violations_ratio
    isi_violations['isi_violations_count'] = isi_violations_count
    isi_violations['isi_violations_of_total_spikes'] = isi_violations_of_total

    return isi_violations



""" Style and aesthetics definitions for plots. """

import pathlib

from itertools import cycle
from functools import wraps

from matplotlib.pyplot import style
import matplotlib.pyplot as plt


###################################################################
###################################################################


def apply_plot_style(style_path=None):
    """Decorator to apply matplotlib style before a plotting function."""

    if style_path is None:
        style_path = pathlib.Path('synapticonn', 'plots', 'plots.mplstyle')

    def decorator_plot_style(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            plt.style.use(style_path)
            return func(*args, **kwargs)
        return wrapper
    return decorator_plot_style


""" file_io.py

File input/output utilities.
"""

import os


###############################################################################
###############################################################################


def fname(file_name, extension):
    """Check a filename, adding an extension if not already specified.

    Parameters
    ----------
    file_name : str
        String that specifies a file name.
    extension : str
        String of the extension (without a period) to be added if one isn't already present.

    Returns
    -------
    file_name : str
        String that specifies a file name.
    """

    if len(file_name.split('.')) == 1:
        file_name = file_name + '.' + extension

    return file_name


def fpath(file_path, file_name):
    """Build the full file path from file name and directory.

    Parameters
    ----------
    file_path : Path or str or None
        Path to the directory where the file is located.
    file_name : str
        Name of the file.

    Returns
    -------
    full_path : str
        Full file path to the file, including directory, if provided.

    Notes
    -----
    This function is mainly used to deal with the case in which file_path is None.
    """

    if not file_path:
        full_path = file_name
    else:
        full_path = os.path.join(file_path, file_name)

    return full_path



""" Custom error definitions. """

class SpikeTimesError(Exception):
    """Base class for custom errors."""

class DataError(SpikeTimesError):
    """Error for if there is a problem with the data."""

class SamplingRateError(SpikeTimesError):
    """Error for if there is a problem with the sampling rate."""

class RecordingLengthError(SpikeTimesError):
    """Error for if there is a problem with the recording length."""

class PlottingError(SpikeTimesError):
    """Error for if there is a problem with plotting."""

class ConnectionTypeError(SpikeTimesError):
    """Error for if there is a problem with the connection type analysis."""
    
class NoDataError(SpikeTimesError):
    """Error for if there is no data to analyze."""
    
class AnalysisError(SpikeTimesError):
    """ Error for if there is a problem with the analysis. """

class SpikePairError(SpikeTimesError):
    """ Error for if there is a problem with the spike pair analysis. """


""" aesthetics.py """


CCG_COLORS = {'pairs': 'blue', 'self': 'green'}

SYN_STRENGTH_COLORS = {'original': 'blue', 'jittered': 'green', 'ci': 'red', 'rectangle': 'lightgrey'}

TIME_UNIT_TIMELAG_LABELS = {'ms': 'Time lag (ms)', 's': 'Time lag (s)'}



""" presence_ratio.py.

Modules for computing presence ratios.
"""

import warnings
import math

import numpy as np

from synapticonn.utils.errors import RecordingLengthError


####################################################
####################################################


def compute_presence_ratio(spike_train,
                           recording_length_t,
                           time_unit,
                           bin_duration_sec=60,
                           mean_fr_ratio_thresh=0.0,
                           srate=None):
    """ Compute the presence ratio of a spike train.

    Parameters
    ----------
    spike_train : numpy.ndarray
        Spike train in milliseconds. If in seconds,
        convert to milliseconds.
    recording_length_t : float
        Length of the recording.
    time_unit : str
        Time unit of the recording.
        Options are 'ms' or 's'.
    bin_duration_sec : float
        Duration of each bin in seconds.
        By default, this is set to 60 seconds.
    mean_fr_ratio_thresh : float
        Minimum mean firing rate ratio threshold.
        This is the minimum mean firing rate that must be present in a bin
        for the unit to be considered "present" in that bin.
        By default, this is set to 0.0. This means that the unit must have
        at least one spike in each bin to be considered "present."
    srate : float
        Sampling rate in Hz.

    Returns
    -------
    presence_ratio : dict
        Dictionary containing the presence ratio.

    Notes
    -----
    Presence ratio is not a standard metric in the field,
    but it's straightforward to calculate and is an easy way to
    identify incomplete units. It measures the fraction of time during a
    session in which a unit is spiking, and ranges from 0 to
    0.99 (an off-by-one error in the calculation ensures
    that it will never reach 1.0).

    Code is adapted from Spike Interface
    (https://github.com/SpikeInterface/spikeinterface/blob/main/src/spikeinterface/qualitymetrics/misc_metrics.py#L1147)

    References
    ----------
    [1] https://spikeinterface.readthedocs.io/en/stable/modules/qualitymetrics/presence_ratio.html
    [2] https://allensdk.readthedocs.io/en/latest/_static/examples/nb/ecephys_quality_metrics.html#ISI-violations 
    """

    # force time unit to seconds
    if time_unit == "s":
        spike_train_sec = spike_train
        recording_length_sec = recording_length_t

    elif time_unit == "ms":
        warnings.warn("Converting spike train and recording length to seconds "
                      "for presence ratio calculation.")
        spike_train_sec = spike_train / 1000
        recording_length_sec = recording_length_t / 1000

    # force float conversion
    mean_fr_ratio_thresh = float(mean_fr_ratio_thresh)

    # validate the input parameters
    _validate_presence_ratio_params(srate, recording_length_sec, bin_duration_sec, mean_fr_ratio_thresh)

    # spike time for unit, in samples
    spike_train_samples = spike_train_sec * srate

    # calculate bin edges
    bin_duration_samples = int((bin_duration_sec * srate))
    total_length = int(recording_length_sec * srate)
    num_bin_edges = total_length // bin_duration_samples + 1
    bin_edges = np.arange(num_bin_edges) * bin_duration_samples

    # calculate thresholds
    unit_fr = len(spike_train_sec) / recording_length_sec
    bin_n_spikes_thresh = math.floor(unit_fr * bin_duration_sec * mean_fr_ratio_thresh)

    # calculate bin edges
    if bin_edges is not None:
        bins = bin_edges
        num_bin_edges = len(bin_edges)
    else:
        bins = num_bin_edges

    h, _ = np.histogram(spike_train_samples, bins=bins)

    presence_ratio = np.sum(h > bin_n_spikes_thresh) / (len(bin_edges) - 1)

    return {"presence_ratio": presence_ratio}


def _validate_presence_ratio_params(srate,
                                    recording_length_t,
                                    bin_duration_sec,
                                    mean_fr_ratio_thresh):
    """ Validate input parameters for presence ratio calculation.

    Parameters
    ----------
    srate : float
        Sampling rate in Hz.
    recording_length_t : float
        Length of the recording in milliseconds.
    bin_duration_sec : float
        Duration of each bin in seconds.
    mean_fr_ratio_thresh : float
        Minimum mean firing rate ratio threshold.
    """

    assert srate > 0, "Sampling rate must be a positive number."
    assert recording_length_t > 0, "Recording length must be greater than 0."
    assert bin_duration_sec > 0, "Bin duration must be greater than 0."

    # --- validate the mean firing rate ratio threshold ---

    if mean_fr_ratio_thresh < 0:
        raise ValueError("The mean firing rate ratio "
                         "threshold must be >= 0.")
    if mean_fr_ratio_thresh > 1:
        warnings.warn("A mean firing rate ratio threshold > 1 may "
                      "lead to low presence ratios.")
    if recording_length_t < bin_duration_sec:
        warnings.warn("Bin duration exceeds recording length. "
                      "Presence ratio will be inaccurate.")

    # --- validate the recording length ---

    # ensure the recording length is greater than the bin duration
    if recording_length_t < bin_duration_sec:
        raise ValueError(f"The recording length of {recording_length_t} "
                         f"sec is shorter than the bin duration of {bin_duration_sec} sec.")

    # --- validate the bin duration ---

    # ensure the bin duration is at least 60 seconds
    if bin_duration_sec < 60:
        warnings.warn("The bin duration is less than 60 seconds."
                      "This may lead to inaccurate presence ratios.")



""" ccg_synaptic_strength.py

Modules for validating synaptic strength using 
cross-correlograms (CCGs).
"""

import warnings
import numpy as np

from joblib import Parallel, delayed

from synapticonn.postprocessing.crosscorrelograms import compute_crosscorrelogram_dual_spiketrains
from synapticonn.utils.errors import SpikeTimesError


##########################################################
##########################################################


def calculate_synaptic_strength(pre_spike_train=None, post_spike_train=None,
                                jitter_range_t=10, num_iterations=1000,
                                max_lag_t=25, bin_size_t=0.5,
                                half_window_t=5, n_jobs=-1):
    """ Calculate the synaptic strength between two spike trains.

    Parameters
    ----------
    pre_spike_train : np.ndarray
        The spike times for the pre-synaptic cell.
    post_spike_train : np.ndarray
        The spike times for the post-synaptic cell.
    jitter_range_t : float
        Jin the range  within which to jitter each spike time.
        Each spike will be shifted by a random amount in the range [-jitter_range_t, +jitter_range_t].
        Default is 10 ms.
    num_iterations : int
        The number of jittered cross-correlograms to compute.
        Default is 1000.
    max_lag_t : float
        The maximum lag to compute the cross-correlogram (in milliseconds).
        Default is 25 ms.
    bin_size_t : float, optional
        The size of each bin in the cross-correlogram (in milliseconds).
        Default is 0.5 ms.
    half_window_t : float, optional
        The half-width of the window around the zero-lag time (in milliseconds).
        Default is 5 ms.
    n_jobs : int, optional
        The number of jobs to run in parallel. Default is -1.

    Returns
    -------
    synaptic_strength_data : dict
        Dictionary containing the original cross-correlogram counts,
        the jittered cross-correlogram counts, the synaptic strength value,
        and the confidence intervals.

    CCG notes
    ---------
    In CCG analysis, if one unit fires before the other, the peak of the CCG
    will be positive. This peak will be centred around the zero-lag time.
    This delay is typically 1-5 ms for excitatory synapses, but can vary
    depending on the type of synapse and the distance between the cells.
    True monosynaptic connections will have a peak at zero lag.

    If the time-lag is less than zero, this indicates that the post-synaptic
    cell is firing before the pre-synaptic cell. In such cases, the direction
    of causality could be reversed, and change depending on which cell is
    considered the pre-synaptic cell. This is why it is important to consider
    the direction of causality when interpreting the CCG results.
    This bidirectional can help confirm the presence of a monosynaptic
    connections.

    Module notes
    ------------
    A single spike train is jittered across multiple iterations to generate
    a distribution of jittered cross-correlograms. This process is repeated
    across a number of iterations to estimate the confidence intervals [1].
    This is introduced to test for the statistical significance of the actual
    cross-correlogram [1].

    It is recommended that the number of iterations be at least 1000 to
    obtain a reliable estimate of the confidence intervals [1].

    The jitter range is recommended to be within a 10 ms range [1].

    Synaptic strength notes
    -----------------------
    If a given unit consistently fires after a second unit, indicated by a peak in the CCG,
    there is high chance that these cells are functionally linked either directly through an excitatory
    synaptic connection or indirectly through a third neuron providing a common input.

    To compute synaptic strength, the firing of a single unit in a pair was jittered across a
    number of iterations (num_iterations) within a time range (jitter_range_t).
    These were used to calculate a confidence interval (CI) between 1% and 99%. If the real
    CCG peak passed the 99% CI, the corresponding functional connection would be considered
    significant and not random.

    A z-score was then performed using the following equation:

    ```Z = x_real - mean_jitter / std_jitter```

    This was used to calculate the synaptic strength value. If the Z-score was greater
    than a positive threshold, the connection was considered significant. This was by
    default set to 5. If the z-score was less than the negative threshold, the connection
    was considered inhibitory. This was by default set to -5.

    References
    ----------
    [1] STAR Protoc. 2024 Jun 21;5(2):103035. doi: 10.1016/j.xpro.2024.103035. Epub 2024 Apr 27
    """

    if (pre_spike_train is None) or (post_spike_train is None):
        raise SpikeTimesError("Pre- and post-synaptic spike trains are required.")

    if bin_size_t > 0.5:
        warnings.warn("Bin size is greater than 0.5 ms. This may affect the accuracy of the synaptic strength value.", UserWarning)
    if max_lag_t > 25:
        warnings.warn("Maximum lag is greater than 25 ms. It is recommended to calculate synaptic strength within a 25 ms range.", UserWarning)
    if jitter_range_t > 10:
        warnings.warn("Jitter range is greater than 10 ms. This may affect the accuracy of the synaptic strength value.", UserWarning)
    if num_iterations < 1000:
        warnings.warn("Number of iterations is less than 1000. This may affect the accuracy of the synaptic strength value.", RuntimeWarning)

    #########################

    synaptic_strength_data = _return_jittered_ccg(pre_spike_train, post_spike_train,
                                                  num_iterations, max_lag_t,
                                                  bin_size_t, jitter_range_t, n_jobs)

    synaptic_strength_data.update(
        _return_synaptic_strength_zscore(synaptic_strength_data['ccg_bins'],
                                         synaptic_strength_data['original_crosscorr_counts'],
                                         synaptic_strength_data['jittered_crosscorr_counts'],
                                         half_window_t, bin_size_t))

    return synaptic_strength_data


def _return_synaptic_strength_zscore(ccg_bins, original_ccg_counts,
                                     jittered_ccg_counts, half_window_t=5,
                                     bin_size_t=0.5):
    """ Calculate the synaptic strength as the Z-score of the peak bin
    count within a specified window in the original CCG.

    Parameters
    ----------
    ccg_bins : np.ndarray
        The time bins for the cross-correlogram.
    original_ccg_counts : np.ndarray
        The original cross-correlogram counts.
    jittered_ccg_counts : np.ndarray
        The jittered cross-correlogram counts.
    half_window_t : float, optional
        The half-width of the window around the zero-lag time (in milliseconds).
        Default is 5 ms.
    bin_size_t : float, optional
        The size of each bin in the cross-correlogram (in milliseconds).
        Default is 0.5 ms.

    Returns
    -------
    synaptic_strength_zscore : dict
        Dictionary containing the synaptic strength value, and the confidence intervals.

    References
    ----------
    [1] STAR Protoc. 2024 Jun 21;5(2):103035. doi: 10.1016/j.xpro.2024.103035. Epub 2024 Apr 27
    """

    assert len(original_ccg_counts) == (jittered_ccg_counts.shape[1]), "Original and jittered CCG counts must have the same length."

    # define the window around the zero-lag time
    window_bins = int((half_window_t*2) / (2 * bin_size_t))
    mid_bin = len(ccg_bins) // 2  # the center bin corresponds to zero lag
    window_slice = slice(mid_bin - window_bins, mid_bin + window_bins + 1)  # slice the window

    # identify the peak bin count within window
    x_real = np.max(original_ccg_counts[window_slice])

    # compute mean and standard deviation of the jittered CCGs
    jittered_window_counts = jittered_ccg_counts[:, window_slice]
    m_jitter = np.mean(jittered_window_counts)
    s_jitter = np.std(jittered_window_counts)

    # calculate Z-score
    if s_jitter > 0:
        synaptic_strength = (x_real - m_jitter) / s_jitter
    else:
        synaptic_strength = np.inf  # if no variance, Z is undefined or infinite

    # calculate confidence intervals within the jittered CCG window
    high_ci = np.percentile(jittered_window_counts, 99, axis=0)
    low_ci = np.percentile(jittered_window_counts, 1, axis=0)

    synaptic_strength_zscore = {'synaptic_strength': synaptic_strength, 'jittered_window_counts': jittered_window_counts,
                                'window_high_ci': high_ci, 'window_low_ci': low_ci, 'window_slice': window_slice}

    return synaptic_strength_zscore


def _return_jittered_ccg(pre_spike_train, post_spike_train, num_iterations=1000,
                         max_lag_t=25, bin_size_t=0.5, jitter_range_t=10, n_jobs=-1):
    """ Return the jittered cross-correlogram.

    Parameters
    ----------
    pre_spike_train : np.ndarray
        The spike times for the pre-synaptic cell.
    post_spike_train : np.ndarray
        The spike times for the post-synaptic cell.
    num_iterations : int
        The number of jittered cross-correlograms to compute.
        Default is 1000.
    max_lag_t : float
        The maximum lag to compute the cross-correlogram (in milliseconds).
        Default is 25 ms.
    bin_size_t : float, optional
        The size of each bin in the cross-correlogram (in milliseconds).
        Default is 0.5 ms.
    n_jobs : int, optional
        The number of jobs to run in parallel. Default is -1.
        If -1, all CPUs are used. If 1, no parallel computing code is used at all,
        which is useful for debugging. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.

    Returns
    -------
    jittered_ccg_data : dict
        Dictionary containing the original cross-correlogram counts and the jittered cross-correlogram counts.

    Notes
    -----
    For reproducibility, a seed is recommended to be set for each iteration.
    This is to ensure that the same jittered spike train is generated for each iteration.
    Hence, the synaptic strength value is consistent across multiple runs.
    """

    assert num_iterations > 2, "Number of iterations must be greater than zero."

    original_ccg_counts, ccg_bins = compute_crosscorrelogram_dual_spiketrains(pre_spike_train, post_spike_train, bin_size_t, max_lag_t)

    # jitter a single spike train across multiple iterations
    # note :: a seed is applied to each iteration for reproducibility
    def single_jitter_iteration(seed):
        jittered_post_spike_train = _apply_jitter(post_spike_train, jitter_range_t, seed=seed)
        jittered_ccg_counts, _ = compute_crosscorrelogram_dual_spiketrains(pre_spike_train, jittered_post_spike_train, bin_size_t, max_lag_t)
        return jittered_ccg_counts

    jittered_ccgs_list = Parallel(n_jobs=n_jobs)(delayed(single_jitter_iteration)(i) for i in range(num_iterations))
    jittered_ccgs = np.vstack(jittered_ccgs_list)

    # calculat confidence intervals on the jittered CCGs
    high_ci = np.percentile(jittered_ccgs, 99, axis=0)
    low_ci = np.percentile(jittered_ccgs, 1, axis=0)

    jittered_ccg_data = {'ccg_bins': ccg_bins, 'original_crosscorr_counts': original_ccg_counts,
                         'jittered_crosscorr_counts': jittered_ccgs, 'high_ci': high_ci, 'low_ci': low_ci}

    return jittered_ccg_data


def _apply_jitter(spike_train, jitter_range_t, seed=None):
    """ Apply random jitter to a spike train within a specified range.

    This is an internal function used to apply random jitter to a spike train.
    It is not recommended to use this function directly.

    Parameters
    ----------
    spike_train : array_like
        The original spike times for a single cell.
    jitter_range_t : float
        The range (in milliseconds) within which to jitter each spike time.
        Each spike will be shifted by a random amount in the range [-jitter_range_t, +jitter_range_t].
    seed : int, optional
        Random seed for reproducibility. Default is 0.

    Returns
    -------
    jittered_spike_train : np.ndarray
        The spike train with added random jitter.

    Notes
    -----
    The output spike train is sorted in ascending order.
    This is to ensure that the jittered spike times are in
    the correct temporal order.

    A seed is recommended to be set for each iteration to ensure
    reproducibility.

    References
    ----------
    [1] https://numpy.org/doc/2.0/reference/random/generated/numpy.random.seed.html 
    """

    assert jitter_range_t > 0, "Jitter range must be greater than zero."

    if seed is not None:
        np.random.seed(seed)  # add seed for reproducibility

    jitter = np.random.uniform(-jitter_range_t, jitter_range_t, size=len(spike_train))
    jittered_spike_train = spike_train + jitter

    sorted_jittered_spike_train = np.sort(jittered_spike_train)  # sort to ensure temporal order

    return sorted_jittered_spike_train



""" test_fit.py

Tests monosynaptic connection fit for the SynaptiConn model object.

NOTES
-----
The tests here are not strong enough to be considered a full test suite.
They serve rather as 'smoke tests', for if anything fails completely.
"""

import synapticonn
import pytest
import pandas as pd

from tests.tutils import load_spiketimes


###############################################


@pytest.mark.parametrize("data_type", ['.mat'])
def test_fit(data_type):
    """ Test the SynaptiConn model object with different spike time data types. """

    # load spike times based on the data type
    spiketimes = load_spiketimes(data_type)

    # ---- test the initialization of the SynaptiConn model ----
    # a .mat file is used for the test
    # with different parameters, to cross-validate the initialization
    # of the SynaptiConn model
    if data_type == '.mat':
        model = synapticonn.SynaptiConn(
            spike_times=spiketimes,
            bin_size_t=0.0005,
            time_unit='s',
            method='cross-correlation',
            max_lag_t=0.20,
            srate=20_000,
            recording_length_t=1000,
        )

    # -- check the spk unit report
    spk_unit_report = model.report_spike_units()
    assert spk_unit_report['n_spikes'][0] == 3238

    # -- check the quality metrics
    params = {'isi_threshold_ms': 1.5,
              'min_isi_ms': 0,
              'presence_ratio_bin_duration_sec': 60,
              'presence_ratio_mean_fr_ratio_thresh': 0.0}

    qc = model.spike_unit_quality(**params)
    presence_ratio = round(float(qc.loc[0, 'presence_ratio']), 2)
    isi_violations_ratio = round(float(qc.loc[0, 'isi_violations_ratio']), 5)
    assert presence_ratio == 1.0
    assert isi_violations_ratio == 0.00267

    # -- model fit
    # check the number of excitatory connections detected
    # w/ set threshold and iterations

    model.set_bin_settings(bin_size_t=0.5, max_lag_t=10, time_unit='ms')
    spike_pairs = [(0, 1), (0, 2), (0, 3), (0, 5), (0, 6)]
    connection_data = model.fit(spike_pairs, synaptic_strength_threshold=5, num_iterations=1000)

    connections_df = pd.DataFrame(connection_data).T
    exc_connections = connections_df[connections_df.putative_exc_connection_type == 'excitatory monosynaptic']
    assert len(exc_connections) == 5



""" info.py

Internal functions to manage info related to the Synapticonn core.
"""


def get_attribute_description():
    """
    Return descriptions of various attributes used in the synaptic connection and spike time analysis.

    Returns
    -------
    attributes : dict
        A dictionary containing descriptions of quality control metrics,
        processing methods, and additional attributes.
    """
    attributes = {
        'qc_metrics': {
            'isi_violations_ratio': (
                'The ratio of the number of interspike interval (ISI) '
                'violations to the total number of spikes.'
            ),
            'isi_violations_rate': 'The rate of interspike interval (ISI) violations.',
            'isi_violations_count': 'The number of interspike interval (ISI) violations.',
            'isi_violations_of_total_spikes': (
                'The ratio of the number of interspike interval (ISI) '
                'violations to the total number of spikes.'
            ),
            'presence_ratio': (
                'The fraction of time during a recording that a unit is spiking '
                'above a defined threshold.'
            ),
            'firing_rate_hz': 'The mean firing rate in Hz (spikes per second).',
            'recording_length_sec': 'The total duration of the recording in seconds.',
            'n_spikes': 'The total number of spikes recorded for a unit.',
        },
        'processing_methods': {
            'cross-correlation': (
                'A measure of similarity between two signals. '
                'Used to identify synaptic connections based on spike train patterns.'
            ),
            'auto-correlation': (
                'A measure of similarity of a signal with a delayed version of itself. '
                'Used to analyze patterns within a single spike train.'
            ),
            'jittering_analysis': (
                'A statistical approach to assess synaptic strength by introducing '
                'controlled temporal perturbations to spike times.'
            ),
            'peak_detection': (
                'Identifies significant peaks in cross-correlograms to estimate '
                'synaptic strength or latency.'
            ),
        },
        'spike_time_attributes': {
            'spike_times': (
                'A dictionary containing spike times for each unit, indexed by unit IDs.'
            ),
            'spike_unit_labels': 'A list of labels representing each spike unit.',
            'time_unit': (
                "The unit of time for spike data. Can be 'ms' (milliseconds) or 's' (seconds)."
            ),
            'recording_length_t': 'The total duration of the recording in the given time unit.',
            'spike_id_type': 'The data type of spike unit identifiers (e.g., int or str).',
            'sampling_rate': 'The frequency at which the recording was sampled, in Hz.',
        },
        'connection_attributes': {
            'synaptic_strength': (
                'A measure of the strength of a monosynaptic connection between two neurons, '
                'typically derived from cross-correlograms.'
            ),
            'connection_types': (
                'Classification of synaptic connections as excitatory, inhibitory, or non-significant '
                'based on thresholds applied to synaptic strength data.'
            ),
            'peak_latency': (
                'The time difference at which the peak of a cross-correlogram occurs, '
                'indicating the latency of a synaptic connection.'
            ),
            'crosscorrelogram': (
                'A histogram of time differences between spikes from two units, '
                'used to analyze their interaction and connection type.'
            ),
        },
        'miscellaneous': {
            'bin_size_t': (
                'The time bin size (in milliseconds) used for computing cross-correlograms or auto-correlograms.'
            ),
            'max_lag_t': (
                'The maximum lag time (in milliseconds) considered when computing cross-correlograms.'
            ),
            'num_iterations': (
                'The number of iterations for statistical analyses like bootstrapping or jittering.'
            ),
            'jitter_range_t': (
                'The range of temporal jitter (in milliseconds) applied to spike trains for analysis.'
            ),
        },
    }
    return attributes


def get_quality_metric_keys():
    """ Return the quality metric keys.

    Returns
    -------
    quality_metric_keys : list
        List of quality metric keys. These 
        are the keys that are used to store
        the quality metrics in the quality
        metric dictionary.
    """

    quality_metric_keys = ['isi_violations_ratio',
                           'isi_violations_count',
                           'isi_violations_of_total_spikes',
                           'presence_ratio',
                           'n_spikes',
                           'firing_rate_hz']

    return quality_metric_keys


def get_available_processing_methods():
    """ Return the available processing methods.

    Returns
    -------
    available_processing_methods : list
        List of available processing methods.
        These are the methods that can be used
        to process the data in the Synapticonn
        core.
    """

    available_processing_methods = ['cross-correlation']

    return available_processing_methods


def get_unit_time_types():
    """ Return the unit time types.

    Returns
    -------
    unit_time_types : list
        List of unit time types. These are
        the types of time units that are
        allowed in the Synapticonn core.
    """

    unit_time_types = ['s', 'ms']

    return unit_time_types



""" core_utils.py

Utility functions for synapticonn package
with reusable utilities.
"""

import pathlib
import logging
import warnings

from functools import wraps
from typing import Any, List, Tuple

import numpy as np

from synapticonn.utils.errors import SpikePairError, SpikeTimesError


######################################################
######################################################


## logging decorator

def setup_log(log_folder_name: str = 'removed_spike_units',
              log_fname: str = 'low_quality_units_removed.log'):
    """ Setup logging for specific class methods.

    Parameters
    ----------
    log_folder_name : str
        Name of the log folder to store the log file.
    log_fname : str
        Name of the log file.
    """

    log_folder = pathlib.Path('logs', log_folder_name)
    log_folder.mkdir(parents=True, exist_ok=True)
    log_path = pathlib.Path(log_folder, log_fname).absolute()
    logging.basicConfig(filename=log_path,
                        format='%(asctime)s - %(levelname)s - %(message)s',
                        level=logging.INFO,
                        force=True)


## helper check decorator


def _validate_parameter(name,
                        value,
                        min_value=None,
                        max_value=None,
                        warn_threshold=None,
                        warn_message=None):
    """ Generic validator for parameters with thresholds and warnings.

    Parameters
    ----------
    name : str
        Name of the parameter.
    value : float
        Value of the parameter.
    min_value : float
        Minimum value of the parameter.
    max_value : float
        Maximum value of the parameter.
    warn_threshold : float
        Warning threshold for the parameter.
    warn_message : str
        Warning message for the parameter.
    """

    if min_value is not None and value <= min_value:
        raise ValueError(f"{name} must be greater than {min_value}.")
    if max_value is not None and value > max_value:
        raise ValueError(f"{name} is greater than the allowed maximum ({max_value}). Adjust the value.")
    if warn_threshold is not None and value > warn_threshold:
        warnings.warn(warn_message, UserWarning)


def _validate_spike_pairs(spike_pairs: List[Tuple] = None,
                          spike_unit_ids: list = None):
    """ Validate spike pairs.

    Check the validity of spike pairs and filter them
    against the available spike unit IDs. These are
    found in the spike unit labels.

    Parameters
    ----------
    spike_pairs : List[Tuple]
        List of spike pairs.
    spike_unit_ids : list
        List of spike unit labels.
        This corresponds with the spike unit IDs
        in the spiketimes data.

    Returns
    -------
    valid_spike_pairs : List[Tuple]
        List of valid spike pairs.
    """

    # ensure spike_unit_ids is a set for efficient lookup
    spike_unit_ids_set = set(spike_unit_ids)

    # validate the input is a list of tuples
    if spike_pairs is not None:
        if not isinstance(spike_pairs, list) or not \
                all(isinstance(pair, tuple) for pair in spike_pairs):
            raise SpikePairError("Spike pairs must be a list of tuples.")
    else:
        raise SpikePairError("Please provide spike pairs to compute synaptic strength.")

    # separate valid and invalid pairs
    valid_spike_pairs = [pair for pair in spike_pairs if pair[0] in \
                         spike_unit_ids_set and pair[1] in spike_unit_ids_set]
    invalid_spike_pairs = [pair for pair in spike_pairs if pair \
                           not in valid_spike_pairs]

    # raise error if no valid spike units
    if not valid_spike_pairs:
        raise SpikeTimesError('No valid spike units to plot.')

    # warn if invalid spike pairs are found
    if invalid_spike_pairs:
        warnings.warn(
            f"Invalid spike pairs found: {invalid_spike_pairs}. These pairs will be ignored.",
            UserWarning
        )

    return valid_spike_pairs



""" acg.py

Modules for plotting autocorrelograms.
"""

import pathlib
import pkg_resources
import matplotlib.pyplot as plt

from synapticonn.postprocessing.autocorrelograms import compute_autocorrelogram
from synapticonn.utils.mod_utils import check_dependency
from synapticonn.plots.plot_utils import acg_ax, check_spktime_ax_length
from synapticonn.plots.style import apply_plot_style
from synapticonn.plots.save import savefig
from synapticonn.plots.aesthetics import TIME_UNIT_TIMELAG_LABELS


##########################################################
##########################################################


style_path = pkg_resources.resource_filename('synapticonn.plots', 'settings.mplstyle')
plt.style.use(style_path)


##########################################################
##########################################################


@savefig
@apply_plot_style(style_path=style_path)
@check_dependency(plt, 'matplotlib')
@check_spktime_ax_length
@acg_ax
def plot_acg(spike_times,
             bin_size_t=1,
             max_lag_t=100,
             show_axes=True,
             ax=None,
             time_unit='ms',
             **kwargs):
    """Plot an autocorrelogram for a single spike train.

    Parameters
    ----------
    spike_times : dict
        Spike times (in milliseconds).
        Each key is a unit ID and each value is a list of spike times.
    bin_size_t : float
        Bin size of the autocorrelogram (in milliseconds).
        Default is 1 ms.
    max_lag_t : float
        Maximum lag to compute the autocorrelogram (in milliseconds).
        Default is 100 ms.
    ax : matplotlib.axes.Axes, optional
        Axis to plot on.
    show_axes: bool, optional
        Whether to add axis labels. Default
        is True.
    time_unit : str, optional
        The time unit for the x-axis. Default is 'ms'.
    **kwargs
        Additional keyword arguments passed to `ax.bar`.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axis with the autocorrelogram plot.
    """

    if not isinstance(spike_times, dict):
        msg = ("Spike train must be a dictionary. "
               "Each key is a unit ID. "
               "Each row is the corresponding spk times.")
        raise ValueError(msg)

    for count, (unit_id, single_spike_times) in enumerate(spike_times.items()):
        lags, autocorr = compute_autocorrelogram(single_spike_times, bin_size_t, max_lag_t)

        ax[count].bar(lags, autocorr, width=bin_size_t, align='center', **kwargs)
        ax[count].set_title(f'Unit {unit_id}')

        if show_axes:
            ax[count].set_xlabel(TIME_UNIT_TIMELAG_LABELS[time_unit])
            ax[count].set_ylabel('Spike counts/bin')

    return ax



""" Sub-module for utility functions. """

from .attribute_checks import requires_recording_length, requires_sampling_rate, requires_arguments
from .report import gen_model_results_str
from .warnings import custom_formatwarning


""" test_initialization.py

Tests for the initialization of the SynaptiConn model object.

NOTES
-----
The tests here are not strong enough to be considered a full test suite.
They serve rather as 'smoke tests', for if anything fails completely.
"""

import synapticonn
import pytest
import numpy as np

from tests.tutils import load_spiketimes


################################


@pytest.mark.parametrize("data_type", ['.mat', 'spikeinterface'])
def test_base_init(data_type):
    """ Test the SynaptiConn model object with different spike time data types. """

    # load spike times based on the data type
    spiketimes = load_spiketimes(data_type)

    # ---- test the initialization of the SynaptiConn model ----
    # a .mat file is used for the test
    # with different parameters, to cross-validate the initialization
    # of the SynaptiConn model
    if data_type == '.mat':
        model = synapticonn.SynaptiConn(
            spike_times=spiketimes,
            bin_size_t=0.0005,
            time_unit='s',
            method='cross-correlation',
            max_lag_t=0.20,
            srate=20_000,
            recording_length_t=1000,
        )

        assert isinstance(model, synapticonn.SynaptiConn)
        assert model.bin_size_t == 0.0005
        assert model.time_unit == 's'
        assert model.max_lag_t == 0.20
        assert model.srate == 20_000
        assert model.recording_length_t == 1000
        assert model.spike_id_type == int

    # a spikeinterface file is used for the test
    # with different time units and parameters to
    # check error handling
    elif data_type == 'spikeinterface':

        si_params = {'spike_times': spiketimes,
                     'bin_size_t': 1000,
                     'time_unit': 'ms',
                     'method': 'cross-correlation',
                     'max_lag_t': 0.20,
                     'srate': 30_000,
                     'recording_length_t': 600*1000,
                     'spike_id_type': np.int32}

        # bin size err
        with pytest.raises(ValueError, match="Bin size is greater than the allowed maximum"):
            model = synapticonn.SynaptiConn(**si_params)

        # time unit err
        si_params['spike_id_type'] = int
        with pytest.raises(synapticonn.utils.errors.SpikeTimesError, match="All keys in spike_times must be of type 'int'. "
                                                                           "Please verify the unit IDs or update the spike_id_type."):
            model = synapticonn.SynaptiConn(**si_params)

        # rec length err
        si_params['spike_id_type'] = np.int32
        si_params['recording_length_t'] = 600
        with pytest.raises(synapticonn.utils.errors.RecordingLengthError, match="Spike times for unit 5 exceed the recording length. Max spike time: 597457.5, "
                                                                                "Recording length: 600. Check that the recording length is correct and in ms."):
            model = synapticonn.SynaptiConn(**si_params)


""" ccg.py

Modules for plotting cross-correlograms.
"""

import pathlib
import pkg_resources
import matplotlib.pyplot as plt

from synapticonn.utils.mod_utils import check_dependency
from synapticonn.plots.style import apply_plot_style
from synapticonn.plots.save import savefig
from synapticonn.plots.plot_utils import ccg_ax, check_spktime_ax_length
from synapticonn.plots.aesthetics import CCG_COLORS, TIME_UNIT_TIMELAG_LABELS


##########################################################
##########################################################



style_path = pkg_resources.resource_filename('synapticonn.plots', 'settings.mplstyle')
plt.style.use(style_path)

##########################################################
##########################################################

@savefig
@apply_plot_style(style_path=style_path)
@check_dependency(plt, 'matplotlib')
@check_spktime_ax_length
@ccg_ax
def plot_ccg(cross_correlograms_data,
             ax=None,
             show_axes=True,
             time_unit='ms',
             **kwargs):
    """
    Plot cross-correlograms for spike train pairs with multiple subplots.

    Parameters
    ----------
    cross_correlograms_data : dict
        Dictionary containing 'cross_correlations' and 'bins' values.
        Can be outputted from `compute_crosscorrelogram` function in `crosscorrelograms.py`.
    ax : numpy.ndarray of matplotlib.axes.Axes
        Array of axes to plot on.
    show_axes : bool, optional
        Whether to add axis labels. Default is True.
    time_unit : str, optional
        Time unit for the x-axis. Default is 'ms'.
    **kwargs
        Additional keyword arguments passed to ax.bar.

    Returns
    -------
    ax : numpy.ndarray of matplotlib.axes.Axes
        Array of axes with the cross-correlogram plots.
    """

    pair_identifiers = (cross_correlograms_data['cross_correllations'].keys())

    # plot the cross-correlograms between all spike-unit pairs
    for count, pair_id in enumerate(pair_identifiers):
        cross_corr = cross_correlograms_data['cross_correllations'][pair_id]
        bins = cross_correlograms_data['bins'][pair_id]
        bin_size_t = bins[1] - bins[0]

        ax[count].bar(bins[:-1], cross_corr, width=bin_size_t, align='center', color=CCG_COLORS['pairs'], **kwargs)

        try:
            pair_id_iter = tuple(pair_id)
        except TypeError:
            pair_id_iter = (pair_id,)

        pair_id_str = ", ".join(map(str, map(int, pair_id_iter)))

        print(f'Pair {pair_id_str} cross-correlogram plotted.')
        ax[count].set_title(f'Pair {pair_id_str}')

        if show_axes:
            ax[count].set_xlabel(TIME_UNIT_TIMELAG_LABELS[time_unit])
            ax[count].set_ylabel('Cross-correlation')

    return ax



""" save.py

Module for saving out figures.
"""

import matplotlib.pyplot as plt
from functools import wraps

from ..utils.file_io import fpath, fname


#############################################
#############################################


def savefig(func):
    """Decorator function to save out figures."""

    @wraps(func)
    def decorated(*args, **kwargs):

        # grab file name and path arguments, if they are in kwargs
        file_name = kwargs.pop('file_name', None)
        file_path = kwargs.pop('file_path', None)
        extension = kwargs.pop('extension', 'png')

        # check for an explicit argument for whether to save figure or not
        #   defaults to saving when file name given (since bool(str)->True; bool(None)->False)
        save_fig = kwargs.pop('save_fig', bool(file_name))

        # check any collect any other plot keywords
        save_kwargs = kwargs.pop('save_kwargs', {})
        save_kwargs.setdefault('bbox_inches', 'tight')

        # check and collect whether to close the plot
        close = kwargs.pop('close', None)

        func(*args, **kwargs)

        if save_fig:
            save_figure(file_name, file_path, close, extension, **save_kwargs)

    return decorated


def save_figure(file_name, file_path=None, close=False, extension='png', **save_kwargs):
    """Save out a figure.

    Parameters
    ----------
    file_name : str
        File name for the figure file to save out.
    file_path : Path or str
        Path for where to save out the figure to.
    close : bool, optional, default: False
        Whether to close the plot after saving.
    extension : str, optional, default: 'png'
        File extension for the figure file to save out.
    save_kwargs
        Additional arguments to pass into the save function.
    """

    full_path = fpath(file_path, fname(file_name, extension))
    plt.savefig(full_path, **save_kwargs)

    if close:
        plt.close()



""" report.py

Utilities for generating reports.

Note that this is a modified version of the report.py file from the neurodsp package
and fooof package, with the original source code available at:
https://github.com/fooof-tools/fooof/blob/main/specparam/core/strings.py#L266
"""

import pandas as pd


###################################################################################################
###################################################################################################

## Settings & Globals
# Centering Value - Long & Short options
# Note: Long CV of 98 is so that the max line length plays nice with notebook rendering
LCV = 98
SCV = 70

###################################################################################################
###################################################################################################


def gen_model_results_str(connection_types, concise, params):
    """Generate a string representation of model monosynaptic connection inference results.

    Parameters
    ----------
    connection_types : dict
        Dictionary of connection types, and the corresponding model results.
    concise : bool, optional, default: False
        Whether to print the report in concise mode.
    params : key-value pairs
        Additional parameters to include in the report
        used for computing the model.

    Returns
    -------
    output : str
        Formatted string of monosynaptic connection results.
    """

    if params.get('method') == 'cross-correlation':

        # params
        ccg_binsize_ms = params.get('bin_size_t')
        max_lag_t = params.get('max_lag_t')
        srate = params.get('srate')
        recording_length_t = params.get('recording_length_t')
        synaptic_strength_threshold = params.get('synaptic_strength_threshold')
        num_iterations = params.get('num_iterations')
        half_window_t = params.get('half_window_t')
        jitter_range_t = params.get('jitter_range_t')
        time_unit = params.get('time_unit')

        # count the number of excitatory and inhibitory connections
        connections = pd.DataFrame(connection_types).T
        exc_connections = connections[connections.putative_exc_connection_type == 'excitatory monosynaptic']
        inh_connections = connections[connections.putative_exc_connection_type == 'inhibitory monosynaptic']

        exc_count = len(exc_connections)
        inh_count = len(inh_connections)

        # identify the pairs
        exc_pairs = exc_connections.index.values
        inh_pairs = inh_connections.index.values

    else:
        raise ValueError('Method not recognized.')

    # create the formatted strings for printing
    str_lst = [

        # header
        '=',
        '',
        'SYNAPTICONN - MONOSYNAPTIC CONNECTIONS',
        '',
        '----------------------------------------',
        # recording parameters
        'Recording Parameters:',
        '',
        'Sampling rate: {} Hz'.format(srate),
        'Recording length: {:1.2f} '.format(recording_length_t) + time_unit,
        '',

        # ccg method parameters
        'Cross-Correlation Method Parameters:',
        '',
        'Crosscorrelogram bin size: {:1.2f} '.format(ccg_binsize_ms) + time_unit,
        'Maximum time lag: {:1.2f} '.format(max_lag_t) + time_unit,
        'Synaptic strength threshold cut-off: {:1.2f}'.format(synaptic_strength_threshold),
        'Half window size used to calculate synaptic strength: {} '.format(half_window_t) + time_unit,
        'Jitter range for synaptic strength computation: {:1.2f} '.format(jitter_range_t) + time_unit,
        'Number of iterations for jitter: {}'.format(num_iterations),
        ''
        '',
        '----------------------------------------',
        '',
    ]

    # add excitatory connection types if exc_pairs exists
    if exc_count > 0:
        str_lst.append(f'{exc_count} excitatory connections were found:')
        str_lst.extend([str(pair) for pair in exc_pairs])
    else:
        str_lst.append('No excitatory connection types were found.')

    str_lst.append('')

    # add inhibitory connections
    if inh_count > 0:
        str_lst.append(f'{inh_count} inhibitory connections were found:')
        str_lst.extend([str(pair) for pair in inh_pairs])
    else:
        str_lst.append('No inhibitory connection types were found.')

    str_lst.append('')

    # footer
    str_lst.extend([
        '',
        '='
    ])

    output = _format(str_lst, concise)

    return output


def _format(str_lst, concise):
    """Format a string for printing.

    Parameters
    ----------
    str_lst : list of str
        List containing all elements for the string, each element representing a line.
    concise : bool, optional, default: False
        Whether to print the report in a concise mode, or not.

    Returns
    -------
    output : str
        Formatted string, ready for printing.
    """

    # Set centering value - use a smaller value if in concise mode
    center_val = SCV if concise else LCV

    # Expand the section markers to full width
    str_lst[0] = str_lst[0] * center_val
    str_lst[-1] = str_lst[-1] * center_val

    # Drop blank lines, if concise
    str_lst = list(filter(lambda x: x != '', str_lst)) if concise else str_lst

    # Convert list to a single string representation, centering each line
    output = '\n'.join([string.center(center_val) for string in str_lst])

    return output



""" correlogram_utils.py

Utilities for correlogram generation.
"""

import numpy as np


##########################################################
##########################################################


def make_bins(max_lag_t, bin_size_t):
    """ Make bins for correlograms.

    Parameters
    ----------
    max_lag_t : float
        Maximum lag to compute the correlograms (in milliseconds).
    bin_size_t : float
        Bin size of the correlograms (in milliseconds).

    Returns
    -------
    bins : array-like
        Bin edges for the correlograms.
    """

    num_bins = int(2 * max_lag_t / bin_size_t) + 1
    bins = np.linspace(-max_lag_t, max_lag_t, num_bins)

    assert len(bins) > 1, "Not enough bins created. Increase max_lag_t or decrease bin_size_t."

    return bins


""" autocorrelograms.py

Modules for generating autocorrelograms.
"""

import numpy as np

from synapticonn.postprocessing.correlogram_utils import make_bins

##########################################################
##########################################################


def compute_autocorrelogram(spike_train, bin_size_t=1, max_lag_t=100):
    """ Compute the autocorrelogram of a spike train.

    Parameters
    ----------
    spike_train : array-like
        Spike times.
    bin_size_t : float, optional
        Bin size of the autocorrelogram.
    max_lag_t : float, optional
        Maximum lag to compute the autocorrelogram.

    Returns
    -------
    lags : array-like
        Lag values.
    autocorr : array-like
        Autocorrelogram values.
    """

    # compute the differences between all spike times
    spike_diffs = np.subtract.outer(spike_train, spike_train)

    # keep only differences within the maximum lag
    spike_diffs = spike_diffs[np.abs(spike_diffs) <= max_lag_t]

    if len(spike_diffs) == 0:
        raise ValueError("No spikes within the maximum lag time."
                         " Please check the spike times and the maximum lag time.")

    # compute histogram (binning the differences with given bin size)
    bins = make_bins(max_lag_t, bin_size_t)
    autocorr, bin_edges = np.histogram(spike_diffs, bins=bins)

    # remove the zero-lag bin (since it's the same spike)
    zero_bin = len(autocorr) // 2
    autocorr[zero_bin] = 0

    # compute lag values centered on each bin
    lags = bin_edges[:-1] + bin_size_t / 2

    return lags, autocorr



""" plot_utils.py

Decorators for plotting utilities.

Notes
------
These should be considered private.
They are not expected to be used outside of this module or used
directly by the user.
"""

import math
import numpy as np
import matplotlib.pyplot as plt

from synapticonn.utils.errors import PlottingError


####################################################
####################################################


def acg_ax(func):
    """ Decorator to check axes for spike-unit labels before plotting multiple subplots.

    Note: this decorator is used for plotting multiple acg subplots. However, 
    to ccg plots, this decorator is not used.
    """
    def wrapper(spike_times, *args, **kwargs):

        n_units = len(spike_times)
        n_cols = min(n_units, 5)  # limit to 5 columns
        n_rows = math.ceil(n_units / n_cols)

        ax = kwargs.get('ax', None)
        figsize = kwargs.get('figsize', (15, 5 * n_rows))

        if ax is None:
            fig, ax = plt.subplots(n_rows, n_cols, figsize=figsize)
            ax = ax.flatten() if isinstance(ax, np.ndarray) else [ax]
        elif ax is not None:
            pass

        kwargs['ax'] = ax
        kwargs.pop('figsize', None)

        return func(spike_times, *args, **kwargs)

    return wrapper


def ccg_ax(func):
    """ Decorator to check axes for spike-unit labels before plotting multiple subplots."""
    def wrapper(cross_correlograms_data, **kwargs):

        ax = kwargs.get('ax', None)
        figsize = kwargs.pop('figsize', (25, 25))

        if ax is None:
            n_pairs = len(list(cross_correlograms_data['cross_correllations'].keys()))
            n_cols = min(n_pairs, 5)  # limit to 5 rows
            n_rows = math.ceil(n_pairs / n_cols)

            _, ax = plt.subplots(n_rows, n_cols, figsize=figsize)
            kwargs['ax'] = ax.flatten() if isinstance(ax, np.ndarray) else [ax]

        return func(cross_correlograms_data, **kwargs)
    return wrapper


def spktrain_ax(func):
    """ Decorator to check axes for spiketrain raster plot. """
    def wrapper(spike_times, *args, **kwargs):

        ax = kwargs.get('ax', None)

        if ax is None:
            figsize = kwargs.pop('figsize', (15, 5)) 
            _, ax = plt.subplots(1, 1, figsize=figsize)

        kwargs['ax'] = ax

        return func(spike_times, *args, **kwargs)

    return wrapper


def check_spktime_ax_length(func):
    """ Decorator to check axes length before plotting multiple subplots. """
    def wrapper(spike_times, *args, **kwargs):

        ax = kwargs.get('ax', None)

        if ax is not None:
            if not isinstance(ax, np.ndarray):  # single axes case
                ax = [ax]
            if len(ax) < len(spike_times):
                msg = ("Number of axes must be equal to the number of units.")
                raise PlottingError(msg)

        kwargs['ax'] = ax

        return func(spike_times, *args, **kwargs)

    return wrapper



""" unit_fr.py

Modules for computing unit firing rates.
"""


####################################################
####################################################


def compute_firing_rates(spike_train, recording_length_t, time_unit):
    """ Compute the firing rates of a spike train.

    Parameters
    ----------
    spike_train : numpy.ndarray (1D)
        Spike train of a single unit.
    recording_length_t : float
        Length of the recording.
    time_unit : str
        Time unit of the recording.
        Options are 'ms' or 's'.

    Returns
    -------
    firing_rates : dict
        Dictionary containing the firing rates
        and the number of spikes.

    Notes
    -----
    Firing rates are calculated as the number of spikes divided by the
    duration of the recording. The firing rate is given in Hz.
    """

    total_spikes = len(spike_train)

    if time_unit == "s":
        unit_fr = total_spikes / recording_length_t
    elif time_unit == "ms":
        unit_fr = total_spikes / (recording_length_t / 1000)
    else:
        raise ValueError("Time unit must be 'ms' or 's'.")

    return {"n_spikes": total_spikes, "firing_rate_hz": unit_fr}


""" tutils.py

Utilities for testing synapticonn."""

import pathlib
import numpy as np
import pandas as pd
import scipy.io

import synapticonn as synapticonn


###############################################################################
###############################################################################


def test_data_path(data_file_type: str) -> pathlib.Path:
    """ Return the path to the test data directory. """

    base_path = pathlib.Path(__file__).parent / "spiketimetest"

    if data_file_type == '.mat':
        # data available for the following reference: 
        # https://star-protocols.cell.com/protocols/3438
        # also found here on the github page:
        # https://github.com/matildebalbilab/STARProtocol_Wangetal2024 
        data_path = pathlib.Path(base_path, "mat_file", "all_unit.mat")
    elif data_file_type == 'spikeinterface':
        data_path = pathlib.Path(base_path, "spikeinterface", "si_spikesorting_array.pkl")
    else:
        raise ValueError(f"Unknown data file type: {data_file_type}")

    return data_path


def load_mat_file(data_path):
    """ Load the .mat file and return the spiketimes. """

    data = scipy.io.loadmat(data_path)
    num_units = len(data['unit_t'][0])

    spiketimes = {}
    for i in range(num_units):
        spiketimes[i] = data['unit_t'][0][i].T[0]

    return spiketimes


def load_spikeinterface_file(data_path):
    """ Load the *.pkl spikeinterface file and return the spiketimes. """

    if data_path.suffix != '.pkl':
        raise ValueError(f"File is not a pickle file: {data_path}")

    data = np.load(data_path, allow_pickle=True)

    spiketimes = data['spike_time_set']

    # convert the spiketimes to milliseconds
    for i in spiketimes.keys():
        spiketimes[i] = (spiketimes[i] / 30000) * 1000

    return spiketimes


def load_spiketimes(data_file_type: str) -> pd.DataFrame:
    """ Load the spiketimes from the test data directory. """

    data_path = test_data_path(data_file_type)

    if data_file_type == '.mat':
        return load_mat_file(data_path)
    elif data_file_type == 'spikeinterface':
        return load_spikeinterface_file(data_path)
    else:
        raise ValueError(f"Unknown data file type: {data_file_type}")



""" data_utils.py 

Module for data utilities with reusable functions.
"""

import numpy as np


######################################################
######################################################


def flatten_list(vals):
    """ Flatten a 2D list to a 1D list.

    Parameters
    ----------
    arr : list
        2D list to be flattened.

    Returns
    -------
    Flattened 1D list.
    """

    if isinstance(vals, list):

        flattened_list = [int(item) for array in vals for item in array]
        unique_list = list(set(flattened_list))

        return unique_list


""" attribute_checks.py

Utils checking attributes of an object if present or not.
"""

import inspect
from ..utils.errors import NoDataError


#######################################################
#######################################################


def requires_sampling_rate(func):
    """ Decorator to ensure that 'srate' (sampling rate) is provided in the instance. """

    def wrapper(self, *args, **kwargs):
        if getattr(self, 'srate', None) is None:
            raise NoDataError('The sampling rate (srate) must be provided.')

        return func(self, *args, **kwargs)

    return wrapper


def requires_recording_length(func):
    """ Decorator to ensure that 'recording_length_t' is provided in the instance. """

    def wrapper(self, *args, **kwargs):
        if getattr(self, 'recording_length_t', None) is None:
            raise NoDataError('The recording length must be provided.')

        return func(self, *args, **kwargs)

    return wrapper


def requires_spike_times(func):
    """ Decorator to ensure that 'spike_times' is provided in the instance. """

    def wrapper(self, *args, **kwargs):
        if getattr(self, 'spike_times', None) is None:
            raise NoDataError('The spike times must be provided.')

        return func(self, *args, **kwargs)

    return wrapper


def requires_arguments(*arg_names):
    """ Decorator to ensure specific arguments are provided. """

    def decorator(func):
        def wrapper(*args, **kwargs):

            # get the function's signature
            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            # check for all required arguments
            missing_args = [arg for arg in arg_names if bound_args.arguments.get(arg) is None]
            if missing_args:
                raise NoDataError(f"The following arguments are required but missing: {', '.join(missing_args)}")
            return func(*args, **kwargs)

        return wrapper
    return decorator


from setuptools import setup, find_packages

def parse_requirements(filename):
    """Read dependencies from a requirements file."""
    with open(filename, "r", encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip() and not line.startswith("#")]

def long_description():
    with open("README.rst", "r", encoding="utf-8") as f:
        return f.read()

if __name__ == "__main__":
    setup(
        name="synapticonn",
        version="0.0.1rc1",
        description="Inferring monosynaptic connections in neural circuits.",
        long_description=long_description(),
        long_description_content_type="text/x-rst",
        author="Michael Zabolocki",
        author_email="mzabolocki@gmail.com",
        maintainer_email="mzabolocki@gmail.com",
        url="https://github.com/mzabolocki/SynaptiConn",
        packages=find_packages(),
        install_requires=parse_requirements("requirements.txt"),
        extras_require={
            "testing": parse_requirements("requirements_dev.txt"),
        },
        python_requires=">=3.8",
        include_package_data=True,
        zip_safe=False,
    )


""" spiketrain.py

Decorators for checking spiketrains before function execution.

Notes
------
These should be considered private.
They are not expected to be used outside of this module or used
directly by the user.
"""

import pathlib
import pkg_resources
import numpy as np
import matplotlib.pyplot as plt

from synapticonn.utils.mod_utils import check_dependency
from synapticonn.plots.style import apply_plot_style
from synapticonn.plots.save import savefig
from synapticonn.plots.plot_utils import spktrain_ax
from synapticonn.plots.aesthetics import TIME_UNIT_TIMELAG_LABELS


##########################################################
##########################################################


style_path = pkg_resources.resource_filename('synapticonn.plots', 'settings.mplstyle')
plt.style.use(style_path)


####################################################
####################################################


@savefig
@apply_plot_style(style_path=style_path)
@check_dependency(plt, 'matplotlib')
@spktrain_ax
def plot_spiketrain(spike_times,
                    ax=None,
                    time_unit='ms',
                    **kwargs):
    """ Plot a spike train.

    Parameters
    ----------
    spike_times : dict
        Spike times.
        Each key is a unit ID and each value is a list of spike times.
    ax : matplotlib.axes.Axes, optional
        Axis to plot on.
    time_unit : str, optional
        Time unit for the x-axis. Default is 'ms'.
    **kwargs
        Additional keyword arguments passed to `ax.eventplot`.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axis with the spike train plot.
    """

    for count, (spk_id, spk_times) in enumerate(spike_times.items()):
        ax.eventplot(spk_times, lineoffsets=count, **kwargs)

    # rename y-axis labels
    ax.set_yticks(np.arange(len(spike_times)))
    ax.set_yticklabels(spike_times.keys())

    # labels
    ax.set_ylabel('Unit ID')
    ax.set_xlabel(TIME_UNIT_TIMELAG_LABELS[time_unit])

    return ax


def check_spiketrain_ndim(func):
    """Decorator to check if array is 1D before function execution.
    To be used on a single spike train (1D), in a list or array format.
    """

    def wrapper(spike_train_ms, *args, **kwargs):
        if len(spike_train_ms) == 0:
            raise ValueError("Array is empty.")
        if spike_train_ms.ndim != 1:
            raise ValueError("Array must be 1D.")
        return func(spike_train_ms, *args, **kwargs)

    return wrapper


def check_spiketrain_millisecond(func):
    """ Decorator to check if array is in milliseconds before function execution.

    To be performed on a single spike train (1D).

    Notes:
    ------
    This assumes that spike times are not in milliseconds if the minimum ISI
    is less than 0,1. Spike ISI from individual neurons should not fire
    faster than 1 ms, so this is a reasonable assumption.
    """

    def wrapper(spike_train_ms, *args, **kwargs):

        min_isi_thresh = 0.1
        msg = ("Check spike times. "
               "If values are not in milliseconds, convert to milliseconds. "
               "Minimum ISI is < 0.1.")

        # flatten spike_train_ms in case it's multi-dimensional
        # handle both 1D and 2D spike trains uniformly
        spike_train_ms = np.atleast_1d(spike_train_ms)

        if spike_train_ms.size == 0:
            raise ValueError("Spike train is empty. Cannot check ISI.")

        # calculate ISI (inter-spike intervals) for the entire array
        isis = np.diff(spike_train_ms, axis=-1)

        # find the minimum ISI across all spike trains
        min_isi = np.min(isis)

        # raise an error if the minimum ISI is below the threshold
        if min_isi < min_isi_thresh:
            raise ValueError(f"{msg} Minimum ISI found: {min_isi} ms.")

        # continue to the wrapped function if no issue
        return func(spike_train_ms, *args, **kwargs)

    return wrapper



""" ccg.py 

Modules for computing cross-correlogram (CCG) features.
"""

import numpy as np 


##################################################################
##################################################################


def compute_ccg_bootstrap(ccg, n_bootstraps=1000):
    """ Compute the bootstrap confidence interval of the CCG.

    Parameters
    ----------
    ccg : array
        Cross-correlogram.
    n_bootstraps : int
        Number of bootstrap samples to draw.

    Returns
    -------
    output : dict
        Dictionary containing the bootstrap standard deviation and the number of bootstraps.
    """

    # arr to store bootstrap samples' means for each iteration
    bootstrap_means = np.zeros(n_bootstraps)

    # perform bootstrap resampling
    for i in range(n_bootstraps):
        bootstrap_sample = np.random.choice(ccg, size=len(ccg), replace=True)
        bootstrap_means[i] = np.mean(bootstrap_sample)

    # calculate confidence intervals (e.g., 95%)
    bootstrap_std = np.std(bootstrap_means)

    return {'bootstrap_std': bootstrap_std}


def compute_ccg_cv(ccg):
    """Compute the coefficient of variation of the CCG.

    Parameters
    ----------
    ccg : array
        Cross-correlogram.

    Returns
    -------
    output : dict
        Dictionary containing the coefficient of variation of the CCG.
    """

    # compute the mean and standard deviation of the CCG
    ccg_mean = np.mean(ccg)
    ccg_std = np.std(ccg)

    # compute the coefficient of variation of the CCG
    ccg_cv = ccg_std / ccg_mean

    return {'ccg_cv': ccg_cv}




""" connections.py

Base model object, which is used to quantify monosynaptic connections between neurons.
"""

import warnings
import logging

import pandas as pd

from typing import Any, List, Tuple

from synapticonn.core.spike_times import SpikeManager
from synapticonn.plots import plot_acg, plot_ccg, plot_ccg_synaptic_strength, plot_spiketrain
from synapticonn.monosynaptic_connections.ccg_synaptic_strength import calculate_synaptic_strength
from synapticonn.monosynaptic_connections.ccg_connection_type import get_putative_connection_type
from synapticonn.postprocessing.crosscorrelograms import compute_crosscorrelogram
from synapticonn.features import compute_peak_latency, compute_ccg_bootstrap, compute_ccg_cv, compute_peak_amp
from synapticonn.core.core_utils import _validate_spike_pairs, _validate_parameter, setup_log
from synapticonn.core.info import get_available_processing_methods
from synapticonn.utils.errors import SpikeTimesError, DataError
from synapticonn.utils.attribute_checks import requires_arguments
from synapticonn.utils.report import gen_model_results_str
from synapticonn.utils.warnings import custom_formatwarning
from synapticonn.utils.data_utils import flatten_list


###############################################################################
###############################################################################

warnings.formatwarning = custom_formatwarning

###############################################################################
###############################################################################


class SynaptiConn(SpikeManager):
    """
    Base class for quantifying monosynaptic connections between neurons.

    This class builds upon the `SpikeManager` to provide functionality for 
    computing, analyzing, and visualizing monosynaptic connections between 
    pairs of neurons based on their spike trains. Key features include 
    cross-correlogram computation, synaptic strength estimation, and connection 
    classification.

    Parameters
    ----------
    spike_times : dict
        Dictionary containing spike times for each unit, indexed by unit ID. Spike
        times should be a NumPy array of floats.
    time_unit : str
        Unit of time for the spike times. Options are 'ms' (milliseconds) or 's' (seconds).
    bin_size_t : float
        Bin size for computing cross-correlograms.
    max_lag_t : float
        Maximum lag to consider when computing cross-correlograms.
    method : str, optional
        Method for computing synaptic strength. Default is 'cross-correlation'.
        Currently, only this method is implemented, but future versions may
        include additional methods.
    recording_length_t : float
        Duration of the recording, in the same time unit as `time_unit`.
    srate : float
        Sampling rate of the recording in Hz.
    spike_id_type : type
        Data type of spike unit identifiers. Typically `int` or `str`.

    Attributes
    ----------
    bin_size_t : float
        Bin size used for cross-correlogram computations.
    max_lag_t : float
        Maximum lag used for cross-correlogram computations.
    method : str
        The method used for synaptic strength computation.
    recording_length_t : float
        Total duration of the recording, in the specified time unit.
    srate : float
        Sampling rate of the spike data, in Hz.
    spike_id_type : type
        Type of spike unit identifiers (e.g., `int` or `str`).
    time_unit : str
        Unit of time used for computations and spike data.
    pair_synaptic_strength : dict
        Dictionary containing synaptic strength data for each spike pair. This 
        is populated after running the `fit` or `synaptic_strength` methods.

    Notes
    -----
    - The `SynaptiConn` class is designed to process and analyze spike train data 
      for determining monosynaptic connections.
    - If the spike times are provided in seconds, they are converted to milliseconds 
      internally to maintain consistency.
    - The class assumes that the recording duration (`recording_length_t`) and 
      time unit (`time_unit`) are accurately specified.

    Examples
    --------
    Initialize the SynaptiConn object and compute synaptic strength:

    >>> synapti_conn = SynaptiConn(
    ...     spike_times=spike_data,
    ...     time_unit='ms',
    ...     bin_size_t=1.0,
    ...     max_lag_t=50.0,
    ...     method='cross-correlation',
    ...     recording_length_t=60000,
    ...     srate=20000
    ... )
    >>> synapti_conn.fit(spike_pairs=[(1, 2), (3, 4)])
    >>> print(synapti_conn.pair_synaptic_strength)
    """

    # connection filtering flag to track if spike pairs have been filtered
    connections_filtering = False


    def __init__(self, spike_times: dict = None,
                 time_unit: str = 'ms',
                 bin_size_t: float = 1,
                 max_lag_t: float = 100,
                 method: str = 'cross-correlation',
                 recording_length_t: float = None,
                 srate: float = None,
                 spike_id_type: type = int):
        """ Initialize the SynaptiConn object. """

        super().__init__(spike_times=spike_times,
                         time_unit=time_unit,
                         srate=srate,
                         recording_length_t=recording_length_t,
                         spike_id_type=spike_id_type)

        self.bin_size_t = self._bin_size_check(bin_size_t, max_lag_t)
        self.max_lag_t = self._max_lag_check(bin_size_t, max_lag_t)
        self.method = self._method_check(method)


    def __call__(self):
        """ Return the object. """

        return self


    def report_correlogram_settings(self):
        """ Report the bin settings. """

        return f"Bin size: {self.bin_size_t} ms, Max lag: {self.max_lag_t} ms"


    def _get_default_settings(self):
        """ Return the settings of the object. """

        settings = {
            'bin_size_t': self.bin_size_t,
            'max_lag_t': self.max_lag_t,
            'method': self.method,
            'recording_length_t': self.recording_length_t,
            'srate': self.srate,
            'spike_id_type': self.spike_id_type,
            'time_unit': self.time_unit
            }

        if self.method == 'cross-correlation':  # default settings for the cross-correlation method

            crosscorr_connection_settings = {
                'bin_size_t': 1,
                'max_lag_t': 100,
                'num_iterations': 1000,
                'jitter_range_t': 10,
                'half_window_t': 5,
                'time_unit': 'ms',
                'n_jobs': -1
            }

            settings.update(crosscorr_connection_settings)

            return settings

        else:
            raise NotImplementedError("Only the 'cross-correlation' method is currently implemented.")


    def spike_unit_ids(self):
        """ Return the spike unit labels. """

        return list(self.spike_times.keys())


    @requires_arguments('bin_size_t', 'max_lag_t', 'time_unit')
    def set_bin_settings(self, bin_size_t: float = None,
                         max_lag_t: float = None,
                         time_unit: str = None,
                         verbose: bool = True):
        """ Set the settings of the object.

        Useful for changing the bin size and maximum lag after initialization.

        Parameters
        ----------
        bin_size_t : float
            Bin size of the cross-correlogram or auto-correlograms.
        max_lag_t : float
            Maximum lag to compute the cross-correlogram.
        time_unit : str
            Time unit options in ms (milliseconds) or s (seconds).
            These are used to set the time unit for the spike times, recording length, 
            bin size, and maximum lag for all processing.
        verbose : bool
            If True, print a message confirming the settings.

        User Warning
        ------------
        Please use this method with caution. Parameters should match the spike time units.
        If a change in the time unit is required, the spike times will NOT be converted to the new unit
        automatically. This should be done manually before calling this method.
        """

        warnings.warn("This method is used to set the bin size and maximum "
                      "lag after initialization. Please use this method with caution. "
                      "Parameters should match the spike time units", UserWarning)

        self.time_unit = self._time_unit_check(time_unit)
        self.bin_size_t = self._bin_size_check(bin_size_t, max_lag_t)
        self.max_lag_t = self._max_lag_check(bin_size_t, max_lag_t)

        if verbose:
            print(f"Bin size set to {self.bin_size_t} {self.time_unit}, "
                  f"and maximum lag set to {self.max_lag_t} {self.time_unit}.")


    @requires_arguments('time_unit')
    def set_time_unit(self, time_unit: str = 'ms'):
        """ Set the time unit for the spike times, recording length, bin size, and maximum lag.

        Parameters
        ----------
        time_unit : str
            Time unit options in ms (milliseconds) or s (seconds).
        """

        warnings.warn("This method is used to set the time unit for the spike times. "
                      "All processing will be based on this time unit.", UserWarning)

        self.time_unit = self._time_unit_check(time_unit)


    def reset_pair_synaptic_strength(self):
        """ Reset the synaptic strength data. """

        if hasattr(self, 'pair_synaptic_strength'):
            del self.pair_synaptic_strength
        else:
            raise DataError("No synaptic strength data found.")


    @requires_arguments('spike_pairs', 'synaptic_strength_threshold')
    def fit(self, spike_pairs: List[Tuple] = None,
            synaptic_strength_threshold: float = 5,
            **kwargs) -> dict:
        """ Compute monosynaptic connections between neurons for a given set of spike times.

        Parameters
        ----------
        spike_pairs : List[Tuple]
            List of spike pairs to compute the synaptic strength.
            These are tuples of pre- and post-synaptic neuron IDs.
            Pre-synaptic neuron ID is the first element and post-synaptic neuron ID is the second element.
        synaptic_strength_threshold : float
            Threshold value for categorizing connection types. Default is 5.
            This is used to categorize the connection types based on the synaptic strength values.
        **kwargs : dict, optional
            Additional parameters for customizing the computation. Includes:
            - num_iterations : int
                Number of iterations for computing synaptic strength (default: 1000).
            - max_lag_t : float
                Maximum lag to compute the synaptic strength (in ms, default: 25.0).
            - bin_size_t : float
                Bin size for computing the synaptic strength (in ms, default: 0.5).
            - jitter_range_t : float
                Jitter range for synaptic strength computation (in ms, default: 10.0).
            - half_window_t : float
                Half window size for synaptic strength computation (in ms, default: 5).
            - n_jobs : int
                Number of parallel jobs to use (default: -1, all cores).

        Attributes set
        --------------
        pair_synaptic_strength : dict
            Dictionary containing the synaptic strength for each pair of neurons.
            This is stored in the object for future reference, and can be accessed using the 'pair_synaptic_strength'
            attribute. This is used to compute the connection types and features, and perform visualizations.

        Raises
        ------
        SpikeTimesError
            If spike pairs are not provided.
        DataError
            If no synaptic strength data is found.
        """

        # compute and set the synaptic strength for the given spike pairs
        synaptic_strength_data = self.synaptic_strength(spike_pairs=spike_pairs, **kwargs)

        # isolate the mono-synaptic connections
        connection_types = self.monosynaptic_connection_types(synaptic_strength_threshold)

        # extract connection features
        # the number of bootstraps can be adjusted, but the default is 1000
        connection_features = self.monosynaptic_connection_features(kwargs.get('n_bootstraps', 1000))

        # merge the connection types and features
        for pair in connection_types:
            connection_types[pair].update(connection_features[pair])

        return connection_types


    def report(self, spike_pairs: List[Tuple] = None,
               synaptic_strength_threshold: float = 5,
               concise: bool = False,
               **kwargs):
        """ Compute the synaptic strength and connection types, and display a report.

        Parameters
        ----------
        spike_pairs : List[Tuple]
            List of spike pairs to compute the synaptic strength.
            These are tuples of pre- and post-synaptic neuron IDs.
            Pre-synaptic neuron ID is the first element and post-synaptic neuron ID is the second element.
        synaptic_strength_threshold : float
            Threshold value for categorizing connection types. Default is 5.
            This is used to categorize the connection types based on the synaptic strength values.
        concise : bool
            If True, print a concise summary of the results. This excludes blank lines.
        **kwargs : dict, optional
            Additional parameters for customizing the computation. Includes:
            - num_iterations : int
                Number of iterations for computing synaptic strength (default: 1000).
            - max_lag_t : float
                Maximum lag to compute the synaptic strength (in ms, default: 25.0).
            - bin_size_t : float
                Bin size for computing the synaptic strength (in ms, default: 0.5).
            - jitter_range_t : float
                Jitter range for synaptic strength computation (in ms, default: 10.0).
            - half_window_t : float
                Half window size for synaptic strength computation (in ms, default: 5).
            - n_jobs : int
                Number of parallel jobs to use (default: -1, all cores).

        Notes
        -----
        Data is computed and displayed in a report format.

        Attributes set
        --------------
        pair_synaptic_strength : dict
            Dictionary containing the synaptic strength for each pair of neurons.
            This is stored in the object for future reference, and can be accessedusing the 'pair_synaptic_strength'
            attribute. This is used to compute the connection types and features, and perform visualizations.
        """

        # find default settings for reporting
        # and update with any additional parameters passed
        settings = {**self._get_default_settings(), **kwargs}
        settings['synaptic_strength_threshold'] = synaptic_strength_threshold

        # compute the synaptic strength and connection types
        connection_types = self.fit(spike_pairs, synaptic_strength_threshold, **kwargs)

        # print the results
        self.print_connection_results(connection_types, concise, settings)


    def print_connection_results(self, connection_types: dict = None,
                                 concise: bool = False,
                                 params: dict = None):
        """ Print the results of the synaptic strength and connection types.

        Parameters
        ----------
        connection_types : dict
            Dictionary containing connection types for all pairs of spike trains.
            This is computed using the 'fit' method.
        concise : bool
            If True, print a concise summary of the results.
            If False, print a detailed summary of the results.
        params : dict
            Additional parameters used for computing the model.
        """

        print(gen_model_results_str(connection_types, concise, params))


    @requires_arguments('spike_pairs', 'num_iterations',
                        'max_lag_t', 'bin_size_t',
                        'jitter_range_t', 'half_window_t')
    def synaptic_strength(self,
                          spike_pairs: List[Tuple] = None,
                          num_iterations: int = 1000,
                          max_lag_t: float = 25.0,
                          bin_size_t: float = 0.5,
                          jitter_range_t: float = 10.0,
                          half_window_t: float = 5,
                          n_jobs: int = -1) -> dict:
        """ Compute the synaptic strength for the given spike pairs.

        Parameters
        ----------
        spike_unit_ids : list
            List of spike unit labels.
        spike_pairs : List[Tuple]
            List of spike pairs to compute synaptic strength.
            These are tuples of pre- and post-synaptic neuron IDs.
            Pre-synaptic neuron ID is the first element and post-synaptic neuron ID is the second element.
        num_iterations : int
            Number of iterations to compute the synaptic strength.
        max_lag_t : float
            Maximum lag to compute the synaptic strength.
        bin_size_t : float
            Bin size of the synaptic strength.
        jitter_range_t : float
            Jitter range to compute the synaptic strength.
        half_window_t : float
            Half window size for the synaptic strength.
        n_jobs: int
            Number of parallel jobs to run. Default is -1.
            Use this to speed up computation.

        Returns
        -------
        synaptic_strength_pairs : dict
            Dictionary containing synaptic strength data for all pairs of spike trains.
            This contains the mean, standard deviation, and confidence intervals of the synaptic strength
            following jittering and bootstrapping.

        Warning
        -------
        If spike times are not in milliseconds, a DataError is raised. Please convert to milliseconds
        for synaptic strength calculations using the 'cross-correlation' method. This is based on [1].

        Attributes set
        --------------
        pair_synaptic_strength : dict
            Dictionary containing the synaptic strength for each pair of neurons.
            This is stored in the object for future reference, and can be accessed using the 'pair_synaptic_strength'
            attribute. This is used to compute the connection types and features, and perform visualizations.

        References
        ----------
        [1] STAR Protoc. 2024 Jun 21;5(2):103035. doi: 10.1016/j.xpro.2024.103035. Epub 2024 Apr 27.

        Notes
        -----
        This method computes the synaptic strength for all pairs of spike trains. Currently, 
        only the cross-correlogram method is implemented. In future versions, this will be expanded
        to include other types of correlation methods, such as cross-correlation, mutual information, etc.

        The 'cross-correlation' method computes the synaptic strength using the cross-correlogram. This method
        performs the following:
            1. a peak detection on the cross-correlogram to estimate the synaptic strength
            2. a statistical analysis to estimate the confidence intervals
            3. a jittering analysis to estimate the jittered synaptic strength.

        Analysis is based on [1]. For excitatory connections, a threshold of 5 is recommended.
        """

        # check spike time units
            # if not in milliseconds, raise an error
            # this is required for synaptic strength calculations
            # please see [1] for more details
        if self.time_unit == 's':
            raise DataError("Spike times are not in milliseconds. Please convert to milliseconds "
                            f"for synaptic strength calculations using {self.method}. "
                            "This can be done by setting the time unit to 'ms' using "
                            "the set_time_unit method.")

        # get spike unit ids
        spike_unit_ids = self.spike_unit_ids()

        # filter passed spike pairs for available spike units
        valid_spike_pairs = _validate_spike_pairs(spike_pairs, spike_unit_ids)

        self.pair_synaptic_strength = {}
        for pre_synaptic_neuron_id, post_synaptic_neuron_id in valid_spike_pairs:

            # retrieve spike times for the pre- and post-synaptic neurons
            pre_synaptic_spktimes = self.get_spike_times_for_units([pre_synaptic_neuron_id]).get(pre_synaptic_neuron_id)
            post_synaptic_spktimes = self.get_spike_times_for_units([post_synaptic_neuron_id]).get(post_synaptic_neuron_id)

            # calculate synaptic strength
            synaptic_strength_data = calculate_synaptic_strength(pre_synaptic_spktimes,
                                                                 post_synaptic_spktimes,
                                                                 jitter_range_t=jitter_range_t,
                                                                 num_iterations=num_iterations,
                                                                 max_lag_t=max_lag_t,
                                                                 bin_size_t=bin_size_t,
                                                                 half_window_t=half_window_t,
                                                                 n_jobs=n_jobs)

            self.pair_synaptic_strength[(pre_synaptic_neuron_id, post_synaptic_neuron_id)] = synaptic_strength_data

        return self.pair_synaptic_strength


    def monosynaptic_connection_types(self, synaptic_strength_threshold: float = None) -> dict:
        """ Categorize monosynaptic connection types based on synaptic strength data output.

        Parameters
        ----------
        synaptic_strength_threshold : float
            Threshold value for categorizing connection types.
            Default is None. This is used to categorize the
            connection types based on the synaptic strength values.

        Returns
        -------
        connection_types : dict
            Dictionary containing connection types for all pairs of spike trains.

        Cross-correlation method notes:
        ------------------------------
        Based on [1], for excitatory connections, a threshold of  5 is recommended.
        For inhibitory connections, a threshold of  -5 is recommended. Thresholds
        can be adjusted based on the synaptic strength data.

        Please see [1] for more details.

        Notes
        -----
        Currently, connection types are based on the synaptic strength values. This is 
        computed using the cross-correlation method. In future versions, this will be expanded
        to include other types of correlation methods.

        References
        ----------
        [1] STAR Protoc. 2024 Jun 21;5(2):103035. doi: 10.1016/j.xpro.2024.103035. Epub 2024 Apr 27.
        """

        if hasattr(self, 'pair_synaptic_strength'):
            connection_types = {}
            for pair, synaptic_strength_data in self.pair_synaptic_strength.items():
                connection_types[pair] = get_putative_connection_type(synaptic_strength_data['synaptic_strength'],
                                                                      threshold=synaptic_strength_threshold)
            return connection_types
        else:
            raise DataError("No synaptic strength data found. Please run the synaptic_strength method first.")


    def monosynaptic_connection_features(self, n_bootstraps: int = 1000) -> dict:
        """ Extract connection features from synaptic strength data.

        Parameters
        ----------
        n_bootstraps : int
            Number of bootstraps to compute the confidence intervals.

        Returns
        -------
        connection_features : dict
            Dictionary containing connection features for all pairs of spike trains.
        """

        if hasattr(self, 'pair_synaptic_strength'):
            connection_features = {}
            for pair, synaptic_strength_data in self.pair_synaptic_strength.items():
                peak_time = compute_peak_latency(synaptic_strength_data['original_crosscorr_counts'], self.bin_size_t)
                peak_amp = compute_peak_amp(synaptic_strength_data['original_crosscorr_counts'])
                std_bootstrap = compute_ccg_bootstrap(synaptic_strength_data['original_crosscorr_counts'], n_bootstraps=n_bootstraps)
                cv_crosscorr = compute_ccg_cv(synaptic_strength_data['original_crosscorr_counts'])

                connection_features[pair] = {'synaptic_strength': synaptic_strength_data['synaptic_strength']}
                connection_features[pair].update(peak_time)
                connection_features[pair].update(peak_amp)
                connection_features[pair].update(std_bootstrap)
                connection_features[pair].update(cv_crosscorr)

            return connection_features
        else:
            raise DataError("No synaptic strength data found. Please run the synaptic_strength method first.")


    def filter_connections(self, connections_df: pd.DataFrame,
                           query: str = None,
                           log: bool = False,
                           overwrite: bool = False) -> pd.DataFrame:
        """ Filter connections based on a query string.

        Parameters
        ----------
        connections_df : pd.DataFrame
            DataFrame containing the connection data.
            This DataFrame should contain the synaptic strength data
            for all select pairs of spike trains.
        query : str
            Query string to filter the DataFrame.
        log : bool
            If True, print the query string used for filtering.
        overwrite : bool
            If True, overwrite the DataFrame with the filtered data.

        Returns
        -------
        connections_df_filtered : pd.DataFrame
            DataFrame containing the filtered connection data.

        Log
        ---
        The log parameter is used to track the removed spike pairs with
        detected monosynaptic connections. This is useful for tracking
        the removed spike pairs from the original DataFrame.

        The output is saved in a log file in the
        'removed_spikepair_connections' folder. The file is named
        'low_quality_connections_removed.log'.
        """

        # validate the input data
        if connections_df is None:
            raise DataError("No connection data found. "
                            "Please run the synaptic_strength method first.")

        # check types
        if not isinstance(connections_df, pd.DataFrame):
            raise TypeError(f"Connections data must be a pandas DataFrame. \
                            Got {type(connections_df)} instead.")
        if not isinstance(query, str):
            raise TypeError(f"Query must be a string. Got {type(query)} instead.")

        # check if the connection data is already filtered
        if SynaptiConn.connections_filtering:
            if not overwrite:
                msg = ("Connections have already been filtered. Please re-initialize the object. "
                       "If this was intentional, please set the 'overwrite' parameter to True.")
                warnings.warn(msg)
            if overwrite:
                SynaptiConn.connections_filtering = False

        # filter based on the query
        if query is not None:
            connections_df_filtered = connections_df.query(query)
        else:
            # skip filtering if no query is provided
            connections_df_filtered = connections_df

        # if log, track removed spike pairs
        # with detected monosynaptic connections
        if log:

            setup_log(log_folder_name='removed_spikepair_connections',
                      log_fname='low_quality_connections_removed.log')

            removed_spikepairs = connections_df[~connections_df.index.isin(connections_df_filtered.index)]

            for key, row in removed_spikepairs.iterrows():
                log_msg = f'spike pair: {key} - removed from original dataframe with query {query}'
                logging.info(log_msg)

        SynaptiConn.connections_filtering = True

        return connections_df_filtered


    def return_crosscorrelogram_data(self, spike_pairs: List[Tuple] = None) -> dict:
        """ Compute and return the cross-correlogram data for valid spike pairs.

        Parameters
        ----------
        spike_pairs : List[Tuple]
            List of spike pairs to compute the cross-correlogram data.

        Returns
        -------
        crosscorrelogram_data : dict
            Dictionary containing cross-correlograms and bins for all pairs of spike trains.
        """

        if self.time_unit == 's':
            warnings.warn("Spike times are not in milliseconds. It is recommended to convert to milliseconds "
                          "for cross-correlogram calculations.", UserWarning)

        # validate spike pairs
        valid_spike_pairs = _validate_spike_pairs(spike_pairs, self.spike_unit_ids())

        # unpack the spike pairs to find spike IDs
        spike_ids = flatten_list(valid_spike_pairs)
        # and retrieve spike times and compute cross-correlogram data
        spike_times = self.get_spike_times_for_units(spike_ids)

        crosscorrelogram_data = compute_crosscorrelogram(
            spike_times, valid_spike_pairs, bin_size_t=self.bin_size_t, max_lag_t=self.max_lag_t)

        return crosscorrelogram_data


    def plot_synaptic_strength(self, spike_pair: tuple = None, **kwargs):
        """ Plot the synaptic strength for the given spike pair.

        Note, this method requires the synaptic strength data to be computed first.
        It only plots the synaptic strength for a single pair of spike trains.

        Parameters
        ----------
        spike_pair : tuple
            Spike pair to plot the synaptic strength.
        **kwargs
            Additional keyword arguments passed to `plot_ccg_synaptic_strength`.

        Raises
        ------
        DataError
            If no synaptic strength data is found.
        SpikePairError
            If spike pair is not provided.
        NotImplementedError
            If the method is not implemented in the current package version.
        """

        if not hasattr(self, 'pair_synaptic_strength'):
            raise DataError("No synaptic strength data found. Please run "
                            "the synaptic_strength method first.")

        # validate spike pair selection
        valid_spike_pair = _validate_spike_pairs([spike_pair], self.spike_unit_ids())

        # check if the method is implemented
        # note that only the 'cross-correlation' method is currently implemented
        # and for future versions, this will be expanded to include other types of correlation methods
        if self.method == 'cross-correlation':
            plot_ccg_synaptic_strength(self.pair_synaptic_strength, valid_spike_pair[0], self.time_unit, **kwargs)
        else:
            raise NotImplementedError("Only the 'cross-correlation' method is currently"
                                      " implemented for plot. Please choose this method.")


    def plot_autocorrelogram(self, spike_units: list = None, **kwargs):
        """ Plot the autocorrelogram.

        Parameters
        ----------
        spike_units : list
            List of spike units to plot.
        **kwargs
            Additional keyword arguments passed to `plot_acg`.

        Notes
        -----
        Autocorrelograms are computed for each spike unit and plotted.
        The bin size and maximum lag are set by the object parameters.
        """

        spike_times = self.get_spike_times_for_units(spike_units)

        if len(spike_times) == 0:
            raise SpikeTimesError("No valid spike units to plot.")

        # plot the autocorrelogram
        plot_acg(spike_times,
                 bin_size_t=self.bin_size_t,
                 max_lag_t=self.max_lag_t,
                 time_unit=self.time_unit,
                 **kwargs)


    def plot_crosscorrelogram(self, spike_pairs: List[Tuple] = None, **kwargs: Any):
        """ Plot the cross-correlogram for valid spike pairs.

        Parameters
        ----------
        spike_pairs : List[Tuple]
            List of spike pairs to plot.
        ax : Any
            Axis to plot on.
        show_axes : bool
            Whether to add axis labels. Default is True.
        **kwargs : Any
            Additional keyword arguments passed to `plot_ccg`.
        """

        crosscorrelogram_data = self.return_crosscorrelogram_data(spike_pairs)
        plot_ccg(crosscorrelogram_data, time_unit=self.time_unit, **kwargs)


    def plot_spike_train(self, spike_units: list = None, **kwargs: Any):
        """ Plot the spike train for the given spike units.

        Parameters
        ----------
        spike_units : list
            List of spike units to plot.
            If None, all spike units are plotted.
        **kwargs : Any
            Additional keyword arguments passed to `plot_spike_train`
            for customizing the plot using `matplotlib`.
        """

        if spike_units is None:
            spktimes_plotting = self.spike_times  # default to all spike times

        elif spike_units is not None:
            try: 
                spktimes_plotting = {k: self.spike_times[k] for k in spike_units}
            except KeyError as e:
                raise SpikeTimesError(f"Please provide valid spike unit keys to plot. "
                                      f"Error with spike units: {e}")

        plot_spiketrain(spktimes_plotting, time_unit=self.time_unit, **kwargs)


    def _bin_size_check(self, bin_size_t, max_lag_t):
        """ Check if the bin size is valid. """

        _validate_parameter(
            name="Bin size",
            value=bin_size_t,
            min_value=0,
            max_value=min(self.recording_length_t, max_lag_t),
            warn_threshold=0.001 if self.time_unit == 's' else 1,
            warn_message="Bin size is greater than 1 ms. This may lead to inaccurate results.",
        )
        return bin_size_t


    def _max_lag_check(self, bin_size_t, max_lag_t):
        """ Check if the maximum lag is valid. """

        _validate_parameter(
            name="Maximum lag",
            value=max_lag_t,
            min_value=0,
            max_value=self.recording_length_t,
        )
        if max_lag_t < bin_size_t:
            raise ValueError("Maximum lag must be greater than or equal to the bin size.")
        return max_lag_t


    def _method_check(self, method):
        """ Check if the method is valid. """

        if method not in get_available_processing_methods():
            raise NotImplementedError(f"Method {method} is not implemented. "
                                      f"Please choose from {get_available_processing_methods()}.")
        return method



""" synaptic_strength_calc.py

Modules for plotting ccg and synaptic strength calculations.
"""

import pathlib
import pkg_resources

import matplotlib.pyplot as plt

from synapticonn.utils.mod_utils import check_dependency
from synapticonn.plots.style import apply_plot_style
from synapticonn.plots.save import savefig
from synapticonn.plots.aesthetics import SYN_STRENGTH_COLORS, TIME_UNIT_TIMELAG_LABELS


##########################################################
##########################################################


style_path = pkg_resources.resource_filename('synapticonn.plots', 'settings.mplstyle')
plt.style.use(style_path)


##########################################################
##########################################################

@savefig
@apply_plot_style(style_path=style_path)
@check_dependency(plt, 'matplotlib')
def plot_ccg_synaptic_strength(pair_synaptic_strength_data,
                               spike_pair,
                               time_unit='ms',
                               ax=None,
                               bbox_to_anchor=(2.2, 1), **kwargs):
    """ Plots synaptic strength based on the provided data using the ccg method.

    Parameters
    ----------
    pair_synaptic_strength_data : dict
        A dictionary containing keys 'ccg_bins', 'original_ccg_counts',
        'jittered_ccg_counts', 'high_ci', 'low_ci', and 'window_slice'.
    spike_pair : tuple
        Spike pair for which synaptic strength is calculated.
    ax : list of matplotlib.axes._subplots.AxesSubplot, optional
        A list of two matplotlib axis objects. If None, new axes will be created.
    time_unit : str, optional
        The time unit for the x-axis. Default is 'ms'.
    bbox_to_anchor : tuple, optional
        The bbox_to_anchor parameter for the legend. Default is (2.2, 1).
    **kwargs : dict
        Additional keyword arguments for customizing the plot.

    Returns
    -------
    ax : list of matplotlib.axes._subplots.AxesSubplot
        A list of two matplotlib axis objects.
    """

    # unpack the data
    data = pair_synaptic_strength_data[spike_pair]
    ccg_bins, ccg_data = data['ccg_bins'], data['original_crosscorr_counts']
    jittered_ccg_counts = data['jittered_crosscorr_counts']
    high_ci, low_ci, window_slice = data['high_ci'], data['low_ci'], data['window_slice']

    if ax is None:
        fig, ax = plt.subplots(1, 2, figsize=(15, 5), sharex=True, sharey=True)
        fig.suptitle(f'Pair: {spike_pair}', fontsize=16)

    # plot the original and jittered ccgs
    ax[0].bar(ccg_bins[1:], ccg_data, width=0.5, color=SYN_STRENGTH_COLORS['original'])
    ax[1].bar(ccg_bins[1:], jittered_ccg_counts.mean(axis=0), width=0.5, color=SYN_STRENGTH_COLORS['jittered'])
    ax[0].plot(ccg_bins[1:], high_ci, color=SYN_STRENGTH_COLORS['ci'], alpha=1, ls='--', label='99% CI')
    ax[0].plot(ccg_bins[1:], low_ci, color=SYN_STRENGTH_COLORS['ci'], alpha=0.3, ls='--', label='1% CI')

    # highlight the selected window used for synaptic strength calculation
    for i in range(2):
        ymin, ymax = ax[i].get_ylim()
        rect = plt.Rectangle((ccg_bins[window_slice.start], 0),
                             ccg_bins[window_slice.stop] - ccg_bins[window_slice.start],
                             ymax, color=SYN_STRENGTH_COLORS['rectangle'], alpha=0.5, label='Selected window')
        ax[i].add_patch(rect)

    # labels
    ax[0].set_title('Original CCG')
    ax[1].set_title('Jittered CCG')

    for i in range(2):
        ax[i].set_xlabel(TIME_UNIT_TIMELAG_LABELS[time_unit])
        ax[i].set_ylabel('Spike count/bin')

    ax[0].legend(bbox_to_anchor=bbox_to_anchor, loc='upper left')

    return ax



""" Sub-module for monosynaptic pair calculations. """

from .ccg_stability import compute_ccg_bootstrap, compute_ccg_cv
from .mono_connection_peak import compute_peak_amp
from .mono_connection_latency import compute_peak_latency